plot.iphula = function(base, path=".") {

    ####################################
    # Load input data from base.params #
    ####################################

    genes = base
    params_name = paste(path, "/", base, ".params", sep="")
    raw_params = scan(params_name, quiet=TRUE, skip=0, what=list("","","",""), sep="\n")

    for (p in 1:length(raw_params)) {
        params_list = strsplit(raw_params[[p]], "\t")
        params = c(params_list[[1]])
        if (params[1] == "rho") {
            rhos = c(params[-1])
        }else{
            if (params[1] == "theta") {
                thetas = c(params[-1])
            }else{
                if (params[1] == "tau") {
                    taus = c(params[-1])
                }else{
                    if (params[1] == "simulations") {
                         sims = as.numeric(c(params[-1]))
                     }
                }
            }
        }
    }


    ##########################
    # Define Plotting set-up #
    ##########################

    pdfname = paste(base, "_all.pdf", sep="")
    pdf(pdfname)

    # Page layout
    n = as.integer(length(rhos))

    if (n>0) {
        if (n%%2 != 0) {
            n = n + 1
        }

        if (n%%2 == 0) {
            if (n%/%2 == 1) {
                mat = matrix(c(1:n),2,1)
            } else {
                if (n%/%2 == 2) {
                    mat = matrix(c(1:n),2,2)
                } else {
                    if (n%/%2 >= 3) {
                        mat = matrix(c(1:6),3,2)
                    }
                }
            }
        }
    }
    else {
        character(0)
    }
    
    layout(mat)
	par(bty="o", lty=1, las=1, ps=8)

    # Color regime of plots
    cm2.colors = function (n) 
    {
        if ((n <- as.integer(n[1])) > 0) {
            even.n <- n%%2 == 0
            k <- n%/%2
            l1 <- k + 1 - even.n
            l2 <- n - k + even.n
            c(if (l1 > 0) hsv(h = 4/6, s = seq(1, ifelse(even.n, 
                1/k, 0), length = l1), v = 1), if (l2 > 1) hsv(h = 0.98, 
                s = seq(0, 1, length = l2)[-1], v = 1))
        }
        else character(0)
    }


    #######################################################
    # First transform matrices to ln likelihood matrices, #
    # output matrices to files "<base>_rho.LNmatrix"      #
    # and find max and min across rhos for graphing       #
    #######################################################

    for (i in 1:length(genes)) {

         collection = numeric()

        # Transform likelihoods in matrices into ln-likelihoods
        for (a in 1:length(rhos)) {
            filename = paste(genes[i], "_", rhos[a], ".matrix", sep="")
            filepath = paste(path, "/", filename, sep="")

            data = read.table(filepath, header=TRUE)
#            data = read.table(filepath, header=FALSE)
            l_matrix = as.matrix(data)
              
            l_matrix[l_matrix==0] = 1/(sims+1)
 
            ln_matrix = log(l_matrix)

            outname = paste(genes[i], "_", rhos[a], ".LNmatrix", sep="")
            outpath = paste(path, "/", outname, sep="")

			no_theta = length(thetas)
            no_tau = length(taus)

            if (as.real(rhos[a]) == 1) {
				no_tau = 1
                write(ln_matrix, file=outpath, ncolumns=no_tau)
            }else{
                write(ln_matrix, file=outpath, ncolumns=no_tau)
            } 

            collection = cbind(collection, ln_matrix)
        }
 
        maxlnlik = as.real(formatC(max(collection), digits = 7), length = 7)
		print(paste("overall max = ", maxlnlik, sep=""))
        minlnlik = as.real(formatC(min(collection), digits = 7), length = 7)


        ################################################
        # Print list of maxima and heat plates per rho #
        ################################################

		listname = paste(genes[i], "_max_list.txt", sep="")
		listpath = paste(path, "/", listname, sep="")
		gene_header = paste(genes[i], sep="")
		write(gene_header, file=listpath)
		list_overall_max = ""

        for (o in 1:length(rhos)) {
			print(paste("rho = ", rhos[o], sep=""))
			rho_header = paste("rho = ", rhos[o], sep="")
			write(rho_header, file=listpath, append = TRUE)
 
            filename = paste(genes[i], "_", rhos[o], ".LNmatrix", sep="")
            filepath = paste(path, "/", filename, sep="")
              
            data = read.table(filepath)
            ln_matrix = as.matrix(data)

            panelmax = max(ln_matrix)
			print(paste("panelmax= ", panelmax, sep=""))
			
			compare_maxlnlik = formatC(maxlnlik, digits = 7)
#			print(paste("compare_maxlnlik= ", compare_maxlnlik))


			# Print list of maxima per rho
			for (te in 1:length(thetas)) {
			
				if (as.real(rhos[o]) == 1) {
					ta = 1
					
#					print(paste("matrix entry=", ln_matrix[te,ta], " panelmax=", panelmax, " comparemax=", compare_maxlnlik))

					if (ln_matrix[te,ta] == panelmax) {
						max_found = paste("   lnLmax = ", panelmax, " at theta = ", thetas[te], " and tau = ", taus[ta], sep="")
						write(max_found, file=listpath, append = TRUE)
					}
					
					if (ln_matrix[te,ta] == compare_maxlnlik) {
						overall_max_found = paste("lnLmax = ", ln_matrix[te,ta], "   for rho = ", rhos[o], ", theta = ", thetas[te], " and tau = ", taus[ta], sep="")
						print(paste("overlall max: ", overall_max_found))
						list_overall_max = paste(list_overall_max, overall_max_found, "\n")
					}
 
				} else {
					for (ta in 1:length(taus)) {
#					print(paste("matrix entry=", ln_matrix[te,ta], " panelmax=", panelmax, " comparemax=", compare_maxlnlik))

						if (ln_matrix[te,ta] == panelmax) {
							max_found = paste("   lnLmax = ", panelmax, " at theta = ", thetas[te], " and tau = ", taus[ta], sep="")
							write(max_found, file=listpath, append = TRUE)
						}
					
						if (ln_matrix[te,ta] == compare_maxlnlik) {
							overall_max_found = paste("lnLmax = ", ln_matrix[te,ta], "   for rho = ", rhos[o], ", theta = ", thetas[te], " and tau = ", taus[ta], sep="")
							print(paste("overlall max: ", overall_max_found))
							list_overall_max = paste(list_overall_max, overall_max_found, "\n")
						}
					}
						
				}
			}
			

			# Print heat panels
            if (as.real(rhos[o]) == 1) {
                tau = 1
 
                # Print into base_all.pdf
                image(seq(1,length(thetas),1), seq(1,length(tau),1), ln_matrix, col=gray((256:0)/256), zlim=(c(minlnlik,maxlnlik)),  main=paste("rho = ", formatC(as.real(rhos[o]), format="fg"), "   max = ", panelmax, sep=""), las=1, xaxt="n", yaxt="n", xlab="Theta", ylab="Tau")
#                image(seq(1,length(thetas),1), tau, ln_matrix, col=cm2.colors(256), zlim=(c(minlnlik,maxlnlik)),  main=paste("rho = ", formatC(as.real(rhos[o]), format="fg"), "   max = ", panelmax, sep=""), las=1, xaxt="n", yaxt="n", xlab="Theta", ylab="Tau")
                axis(side=1, at=seq(1,length(thetas),1), labels=formatC(as.real(thetas), format="fg"))

				# Print individual base_rho.pdf
				pdfname_rho = paste(genes[i], "_", rhos[o], ".pdf", sep="")
				pdf(pdfname_rho)
                image(seq(1,length(thetas),1), seq(1,length(tau),1), ln_matrix, col=gray((256:0)/256), zlim=(c(minlnlik,maxlnlik)),  main=paste("rho = ", formatC(as.real(rhos[o]), format="fg"), "   max = ", panelmax, sep=""), las=1, xaxt="n", yaxt="n", xlab="Theta", ylab="Tau")
#                image(seq(1,length(thetas),1), tau, ln_matrix, col=cm2.colors(256), zlim=(c(minlnlik,maxlnlik)),  main=paste("rho = ", formatC(as.real(rhos[o]), format="fg"), "   max = ", panelmax, sep=""), las=1, xaxt="n", yaxt="n", xlab="Theta", ylab="Tau")
                axis(side=1, at=seq(1,length(thetas),1), labels=formatC(as.real(thetas), format="fg"))
				dev.off()
					
            } else {
                # Print into base_all.pdf
                image(seq(1,length(thetas),1), seq(1,length(taus),1), ln_matrix, col=gray((256:0)/256), zlim=(c(minlnlik,maxlnlik)), main=paste("rho = ", formatC(as.real(rhos[o]), format="fg"), "   max = ", panelmax, sep=""), las=1, xaxt="n", yaxt="n", xlab="Theta", ylab="Tau")
#                image(seq(1,length(thetas),1), seq(1,length(taus),1), ln_matrix, col=cm2.colors(256), zlim=(c(minlnlik,maxlnlik)), main=paste("rho = ", formatC(as.real(rhos[o]), format="fg"), "   max = ", maxlnlik, sep=""), las=1, xaxt="n", yaxt="n", xlab="Theta", ylab="Tau")
                axis(1, at=seq(1,length(thetas),1), labels=formatC(as.real(thetas), format="fg"))
                axis(2, at=seq(1,length(taus),1), labels=formatC(as.real(taus), format="fg"), las=1)

				# Print individual base_rho.pdf
				pdfname_rho = paste(genes[i], "_", rhos[o], ".pdf", sep="")
				pdf(pdfname_rho)
                image(seq(1,length(thetas),1), seq(1,length(taus),1), ln_matrix, col=gray((256:0)/256), zlim=(c(minlnlik,maxlnlik)), main=paste("rho = ", formatC(as.real(rhos[o]), format="fg"), "   max = ", panelmax, sep=""), las=1, xaxt="n", yaxt="n", xlab="Theta", ylab="Tau")
#                image(seq(1,length(thetas),1), seq(1,length(taus),1), ln_matrix, col=cm2.colors(256), zlim=(c(minlnlik,maxlnlik)), main=paste("rho = ", formatC(as.real(rhos[o]), format="fg"), "   max = ", maxlnlik, sep=""), las=1, xaxt="n", yaxt="n", xlab="Theta", ylab="Tau")
                axis(1, at=seq(1,length(thetas),1), labels=formatC(as.real(thetas), format="fg"))
                axis(2, at=seq(1,length(taus),1), labels=formatC(as.real(taus), format="fg"), las=1)
				dev.off()
					
            }
			
        } # end of rho loop
		
		overall_text = paste("\nMaximum over all parameters:\n", list_overall_max, sep="")
		write(overall_text, file=listpath, append = TRUE)

    } # end of genes loop

    dev.off()

    print(paste("PDF file ", pdfname, " was produced."))
    print("plot.iphula finished.")

} # end function plot.iphula

#done
