/*
 * iphula.c
 *
 *
 * Part of (p)IPHULA 1.16 (May 2007)
 *
 * (c) 2004-2007 by Heiko A. Schmidt, Jutta Buschbom, 
 *                  Gunter Weiss, and Arndt von Haeseler
 * (c) 1998-2004 by Gunter Weiss and Arndt von Haeseler
 *
 * All parts of the source except where indicated are distributed 
 * under the GNU public licence 2. 
 * See http://www.opensource.org/licenses/gpl-license.php details.
 *
 * ($Id: iphula.c,v 1.16 2005/09/30 01:23:17 hschmidt Exp $)
 *
 */


#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <string.h>

#define NRANSI
#define EXTERN

#include "iphula.h"
#include "sub1.h"
#include "sub2.h"
#include "config.h"

#if 0
#   define VERBOSE 1
#   define FIXEDRAND 1
#endif

#ifdef _OPENMP
#	define USE_OPENMP 1
#	include "omp.h"
#else
#	ifdef USE_OPENMP
#		undef USE_OPENMP
#	endif
#endif

#ifdef MPI
#	define USE_MPI 1
#	include <mpi.h>
#else
#	ifdef USE_MPI
#		undef USE_MPI
#	endif
#endif

#define SIMPERDOT 100

 /* variables  */

long randseed;                   /* randomized start  */

double factor;
double Tt;

/* double *observed; */	/* not really used - delete? (HAS) */

int NumVarPos;
double MeanPWDist;
int NumThreads  = 1;	/* number of threads (1, if no OpenMP) (HAS) */
int MyThreadNum = 0;	/* thread number (0, if no OpenMP) (HAS) */
int SetNumThreads = -1;	/* user-set number of threads (-1, if no OpenMP) (HAS) */


/*** TODO: all float -> double  - DONE (20050927) ***/

double ***prob_com;		/* TODO: ask Gunter: why float and not uli or double ? (HAS) */

double tolerance, thetaScale;

/*TTree *tree;
TNode **Tlist;*/

FILE *prob_fp;
FILE *matr_fp;
FILE *params_fp;

/***************************************************************/
#ifdef USE_MPI
#	include "mpislavequeue.c"
#	include "mpiiphula.c"
#endif
/***************************************************************/

#if 0
void iph_exit(int exitcode, char *exitstr, int printline, char *fname, int fline, int dowait)
{
        if(exitstr != NULL) {
                fprintf(stderr, "%s", exitstr);
                if(printline) fprintf(stderr, " (%s:%d)", fname, fline);
                fprintf(stderr, "\n");
        }
#       ifdef USE_MPI
                PI_Finalize();
#       endif

        if(dowait) {
                printf("\npress [return] to finish:");
                fflush(stdout);
                while (getchar() != '\n');
        }
        exit(exitcode);
} /* iph_exit */
#       endif


/***************************************************************/


void printusage(char *fname)
{
                 /* |<---                      80 characters                                   --->| */
   /* fprintf(stderr, "\n\nUsage: %s [-h] [options] [ Infilename [ UserTreeFilename ] ]\n\n", fname); */
   fprintf(stderr, "\n\nUsage: %s [-h] [options]\n\n", fname);
   fprintf(stderr, "  -h               - print usage\n");
   fprintf(stderr, "  -prefix=XXX      - use 'XXX' as filename prefix instead of 'panel'\n");
   fprintf(stderr, "                     or 'INFILE'\n");
   fprintf(stderr, "  -numthread<#>    or\n");
   fprintf(stderr, "  -numthread=<#>   - use <#> OpenMP threads (default=#CPUs)\n");
   fprintf(stderr, "  -randseed<#>     or\n");
   fprintf(stderr, "  -randseed=<#>    - use <#> as random number seed, for debug purposes only\n");
   fprintf(stderr, "  -raw             - write raw simulation data to *.raw\n");
   fprintf(stderr, "  -infile=INFILE   - use file INFILE for data input instead of stdin\n");
   fprintf(stderr, "\n");
#  ifdef USE_MPI
#    ifdef PVERBOSE
	fprintf(stderr,"[%2d] sending done1\n", PI_Myid);
#    endif /* PVERBOSE */
        PI_SendDone();
        MPI_Finalize();
#  endif 
   exit (1);
#        if 0
   iph_exit(1, NULL, FALSE, __FILE__, __LINE__, exit_wait_optn);
#        endif
} /* printusage */

/***************************************************************/

void scancmdline(int *argc, char **argv[])
{
   short flagused;
   int n;
#if 0
   int found;                           /* "click" found in cmd name? */
#endif
   int count, dummyint;

   /* set defaults */
   sprintf(PANELFILE,  "%s.%s", PREFIXDEFAULT, PANELFILEEXT);
   sprintf(PARAMSFILE, "%s.%s", PREFIXDEFAULT, PARAMSFILEEXT);
   /* sprintf(PARAMSFILE, "%s.%s", FILEPREFIX, PARAMSFILEEXT); */
   sprintf(RAWFILE,    "%s.%s", PREFIXDEFAULT, RAWFILEEXT);
   setinfile_optn = SETINFILE_NONE;
   setprefix_optn = SETPREFIX_NONE;
   writeraw_optn  = FALSE;


#if 0
   double dummydbl;
   found  = FALSE;
#endif

   
   for (n = 1; n < *argc; n++) {
#     ifdef VERBOSE1
         printf("argv[%d] = %s\n", n, (*argv)[n]);
#     endif

      flagused = FALSE;

      /* use given prefix rather than INFILENAME for output filename */
      dummyint = 0;
      count = sscanf((*argv)[n],"-prefix=%s%n", FILEPREFIX, &dummyint);
      if (dummyint>8) {
        printf("PREFIX set to = \"%s\"\n", FILEPREFIX);
        setprefix_optn = SETPREFIX_PREFIX;
        flagused = TRUE;
      }

      /* use given prefix rather than INFILENAME for output filename */
      dummyint = 0;
      count = sscanf((*argv)[n],"-infile=%s%n", INFILENAME, &dummyint);
      if (dummyint>8) {
        printf("INFILENAME set to = \"%s\"\n", INFILENAME);
       	setinfile_optn = SETINFILE_INFILE;
        if (setprefix_optn == SETPREFIX_NONE) {
        	strcpy(FILEPREFIX, PREFIXDEFAULT);
        	/* strcpy(FILEPREFIX, INFILENAME); */
        	printf("PREFIX set to = \"%s\"\n", FILEPREFIX);
        	setprefix_optn = SETPREFIX_PREFIX;
	}
        flagused = TRUE;
      }

#if 0
      /* print version */
      dummyint = 0;
      count = sscanf((*argv)[n],"-V%n", &dummyint);
      if (dummyint==2) {
        printversion((*argv)[0]);
        flagused = TRUE;
      }

      /* print version */
      dummyint = 0;
      count = sscanf((*argv)[n],"-version%n", &dummyint);
      if (dummyint==8) {
        printversion((*argv)[0]);
        flagused = TRUE;
      }
#endif

      /* print help message */
      dummyint = 0;
      count = sscanf((*argv)[n],"-h%n", &dummyint);
      if (dummyint==2) {
        printusage((*argv)[0]);
        flagused = TRUE;
      }

      /* write raw data to *.raw */
      dummyint = 0;
      count = sscanf((*argv)[n],"-raw%n", &dummyint);
      if (dummyint==4) {
        writeraw_optn = TRUE;
        flagused = TRUE;
      }

      /* set random number seed to fixed value */
      count = sscanf((*argv)[n],"-randseed%d", &dummyint);
      if (count == 1) {
        randseed = dummyint;
        flagused = TRUE;
      }

      /* set random number seed to fixed value */
      count = sscanf((*argv)[n],"-randseed=%d", &dummyint);
      if (count == 1) {
        randseed = dummyint;
        flagused = TRUE;
      }

      /* set random number seed to fixed value */
      count = sscanf((*argv)[n],"-numthread%d", &dummyint);
      if (count == 1) {
	SetNumThreads = dummyint;
        flagused = TRUE;
      }

      /* set random number seed to fixed value */
      count = sscanf((*argv)[n],"-numthread=%d", &dummyint);
      if (count == 1) {
	SetNumThreads = dummyint;
        flagused = TRUE;
      }


      if (flagused == FALSE) {
         fprintf(stderr, "WARNING: commandline parameter %d not recognized (\"%s\")\n", n, (*argv)[n]);
      }
      flagused = FALSE;
   } /* while argc */

   /* setprefix_optn = SETPREFIX_INFILE; */
   if (setprefix_optn != SETPREFIX_NONE) {
   	sprintf(PANELFILE,  "%s.%s", FILEPREFIX, PANELFILEEXT);
   	sprintf(PARAMSFILE, "%s.%s", FILEPREFIX, PARAMSFILEEXT);
   	sprintf(RAWFILE,    "%s.%s", FILEPREFIX, RAWFILEEXT);
   } else {
        strcpy(FILEPREFIX, PREFIXDEFAULT);
        /* strcpy(FILEPREFIX, INFILENAME); */
        printf("PREFIX set to = \"%s\"\n", FILEPREFIX);
        setprefix_optn = SETPREFIX_PREFIX;
   }

} /* scancmdline */


/***************************************************************/

#ifdef USE_MPI
int slave_loop(FILE *rawfile)
/* int slave_loop(int argc, char *argv[], FILE *rawfile) */
{
	int        notdone;
	MPI_Status stat;
	uli        simul;
	double theta0;
	double Rho;
#if 0
	long counter;
	long tmp;
#endif
	unsigned long int num_positive;	/* TODO: ask Gunter: double/uli ? (HAS) */
	TNode *T1, *T2;

#	if 0
	PI_RecvParams(
		Freq,		/* SubsProDialog */
		&kappa,		/* SubsProDialog */
		&ksi,		/* SubsProDialog */
		&NumCats,	/* SubsProDialog */
		&GammaShape,	/* SubsProDialog */
		&NumSamples,	/* DataDialog */
		&NumBases,	/* DataDialog */
		&NumVarPos,	/* DataDialog */
		&MeanPWDist,	/* DataDialog */
/* change ;-) - done */
		/* &Rho, */	/* SimParDialog */
		&Rho_arr,	/* SimParDialog */
		&MinTau,	/* SimParDialog */
		&MaxTau,	/* SimParDialog */
		&Tausteps,	/* SimParDialog */
		&MinTheta,	/* SimParDialog */
		&MaxTheta,	/* SimParDialog */
		&Thetasteps,	/* SimParDialog */
		&NumRuns,	/* SimParDialog */
		&tolerance,	/* SimParDialog */
		&randseed);	/* random number seed */

	initrandom(randseed);
#	endif

#	ifdef VERBOSE
		fprintf(stderr,"[%2d] entering slave loop!\n", PI_Myid);
#	endif
#	ifdef PVERBOSE_RECV_START
		fprintf(stderr, "[%2d] Probing for work\n", PI_Myid);
#	endif /* PVERBOSE_RECV_START */

	MPI_Probe(PI_MyMaster, MPI_ANY_TAG, PI_Comm, &stat);

#	ifdef PVERBOSE_RECV_END
		fprintf(stderr, "[%2d] ... received TAG=%d\n", PI_Myid, stat.MPI_TAG);
#	endif /* PVERBOSE_RECV_END */

	notdone = (stat.MPI_TAG != PI_DONE);
	if (!notdone) {
#		ifdef VERBOSE
			fprintf(stderr, "[%d] stopping...\n", PI_Myid);
#		endif
		PI_RecvDone();
	}

	while (notdone) {
		switch (stat.MPI_TAG) {
			case PI_DOSIMUL: {
				uli r;
				/* Receive specs for simulation */
/* change :-) - no change necessary, single Rho - done */
				PI_RecvDoSimulBlock(&simul, &theta0, &Tt, &Rho);
				num_positive = 0;
/* change :-) - OK, due to single Rho - done */
				factor = 1.0/Rho;

#				ifdef VERBOSE
					fprintf(stderr, "[%2d] doing %lu simulations\n", PI_Myid, simul);
#				endif

				for (r = 1; r <= simul; r++) {
					/* do something */
#if 0
#					pragma omp critical
					{ counter++; tmp = counter; }
#endif

					/* cunstruct random tree */
					/* grow_tree(factor,Tt); */		/* HAS TO BE THREAD-SAFE (HAS) */
					grow_tree(factor, Tt, &tree, &Tlist);

					/* simulate */
					seq_gen(tree,theta0,&num_positive, rawfile);	/* HAS TO BE THREAD-SAFE (HAS) */

					/* clean up */
					T1=tree->nodeList;
					while(T1 != NULL) {
						T2=T1;
						T1=T1->next;
						free(T2);
					}
					free(Tlist);
				} /* end of loop */

				/* send results */
				PI_SendSimulBlock(num_positive, simul);
				break;
				}
			default: {
				fprintf(stderr, "ERROR: Unknown Message Tag %d received\n", stat.MPI_TAG);
				MPI_Abort(PI_Comm, 1);
				}
		} /* switch stat.MPI_TAG */ 

#		ifdef PVERBOSE_RECV_START
			fprintf(stderr, "[%2d] Probing for work\n", PI_Myid);
#		endif /* PVERBOSE_RECV_START */

		MPI_Probe(PI_MyMaster, MPI_ANY_TAG, PI_Comm, &stat);

#		ifdef PVERBOSE_RECV_END
			fprintf(stderr, "[%2d] ... received TAG=%d\n", PI_Myid, stat.MPI_TAG);
#		endif /* PVERBOSE_RECV_END */

		notdone = (stat.MPI_TAG != PI_DONE);
		if (!notdone)
			PI_RecvDone();

	} /* while notdone */

	return	(0);
} /* slave_loop*/
#endif /* USE_MPI */



/***************************************************************/


int main(int argc, char *argv[])
{
	time_t zeit1;
	time_t zeit2;
#	ifdef VERBOSE
#	endif

	FILE *infile;
	FILE *rawfile = NULL;
	unsigned long int num_positive;	/* TODO: ask Gunter: double/uli ? (HAS) */
	long counter;

        double numlength=3;

#ifdef USE_OPENMP
#endif /* USE_OPENMP */
#ifndef USE_MPI
	long tmp;
	long r;
#endif /* !USE_MPI */

	int i, f, n;
	/* int m, j, f_min, f_max; (HAS) */
	int d;
	/* int N_0, N_0_stepsize, N_0_step; (HAS) */
	int tt;
	int Taustepscount;
	/* int Tx, Tx_stepsize, Tx_step; (HAS) */
	
	double theta0;
	double Rho;
	double Tau;
	/* double gen; */  /* never really used !!! delete? (HAS) */
	/* double mY; */  /* never really used !!! delete? (HAS) */
	/* double *factor_v; (HAS) */

#ifndef USE_MPI
	TNode *T1, *T2;
#endif /* !USE_MPI */

#	ifdef USE_OPENMP
		TTree  **alltrees;
		TNode ***allnodes;
#	endif

	/* default settings */

#	if 0
		NumCats = 8;		/* rate heterogeneity */
		GammaShape = 0.24;
#	endif

	NumCats = 1;		/* rate heterogeneity */
	GammaShape = 0.0;

	/* character frequencies */
	/* Freq[0] = 0.327; */
	/* Freq[1] = 0.338; */
	/* Freq[2] = 0.114; */
	/* Freq[3] = 0.221; */

	Freq[0] = 0.250;	/* character frequencies */
	Freq[1] = 0.250;
	Freq[2] = 0.250;
	Freq[3] = 0.250;

	/* kappa=15.5; */
	/* ksi=1.85; */

	kappa=0.5;
	ksi=1.00;

	MinTheta = 1.0;		/* Theta in 10 steps [1 .. 10] */
	MaxTheta = 10.0;
	Thetasteps = 10;

	MinTau = 0.5;		/* Tau in 10 steps [0.5 .. 5] */
	MaxTau = 5;
	Tausteps = 10;

	Rho_arr[0]=0.001;	/* Rho in 8 exponetial steps [10^-2 .. 10^4] */
	Rho_arr[1]=0.01;
	Rho_arr[2]=0.1;
	Rho_arr[3]=1;
	Rho_arr[4]=10;
	Rho_arr[5]=100;
	Rho_arr[6]=1000;
	Rho_arr[7]=10000;
	Rhosteps = 8;

	tolerance = 0.2;

	/* Opening Monolog */
	OpenDialog();

#	ifdef USE_MPI
		PI_Init(&argc, &argv);
		if (PI_IamMaster) {
#	endif /* USE_MPI */
			/* Get Options fron Cmdline */
			scancmdline(&argc, &argv);

			/* open output file "panel" */
			if (setinfile_optn == SETINFILE_INFILE) {
				infile = fopen(INFILENAME,"r");
				if (infile == NULL) {
					fprintf(stderr,"Error opening input file '%s': reading from stdin\n", INFILENAME);
					infile = stdin;
				}
			} else {
				infile = stdin;
			}


			/* open output file "panel" */
			if ( (prob_fp = fopen(PANELFILE,"w")) == NULL) {
				fprintf(stderr,"ERROR: cannot open outputfile \"%s\"!", PANELFILE);
				exit(1);
			}

			/* print header to "panel" file */
#			ifdef USE_MPI
#				ifdef USE_OPENMP
					fprintf(prob_fp,"\nOutfile of your IPHULA run (%s, hybrid MPI/OpenMP binary)\n", VERSION); fflush(prob_fp);
#				else
					fprintf(prob_fp,"\nOutfile of your IPHULA run (%s, MPI-parallel binary)\n", VERSION); fflush(prob_fp);
#				endif
#			else
#				ifdef USE_OPENMP
					fprintf(prob_fp,"\nOutfile of your IPHULA run (%s, OpenMP-parallel binary)\n", VERSION); fflush(prob_fp);
#				else
					fprintf(prob_fp,"\nOutfile of your IPHULA run (%s, sequential binary)\n", VERSION); fflush(prob_fp);
#				endif
#			endif
			fprintf(prob_fp,"\n****************************************\n"); fflush(prob_fp);
			{
				struct tm *tmptime; char tmpchar[1000];
				tmptime = localtime(&zeit1);
				strcpy(tmpchar, asctime(tmptime));
				fprintf(prob_fp,"%s", tmpchar); fflush(prob_fp);
				/* fprintf(prob_fp,"%s", asctime(localtime(&zeit1))); fflush(prob_fp); */
			}
#			ifdef USE_MPI
				/* SetNumThreads = 2; */
				fprintf(prob_fp,"MPI parallel with %d processes (ID: %d)\n", PI_NumProcs, PI_Myid);
#			endif /* USE_MPI */
#if 0
#			ifdef USE_OPENMP
				/* SetNumThreads = 2; */
				if (SetNumThreads > 0)
					omp_set_num_threads(SetNumThreads);
				NumThreads  = omp_get_max_threads();
				MyThreadNum = omp_get_thread_num();
				fprintf(prob_fp,"OpenMP parallel with %d threads (max %d)\n", NumThreads, omp_get_max_threads());
#			endif
#endif
#			ifdef FIXEDRAND
				fprintf(prob_fp, "RANDOM NUMBER SEED FIXED TO -1 FOR DEBUGGING!!!\n");
#			endif

		       	fprintf(prob_fp,"Run started with commandline: ");
		      	for (n = 0; n < argc; n++)
				fprintf(prob_fp,"%s ",argv[n]);
			fprintf(prob_fp,"\n");

			/* Opening Monolog */
			/* OpenDialog(); */ /* moved further up */
			/* setting: Freq[0..3], kappa(ts/tv ratio), ksi(pyr/pur ts ratio), GammaShape */
			SubsProDialog(infile, prob_fp);
			/* setting: NumSamples, NumBases, NumVarPos, MeanPWDist */
			DataDialog(infile, prob_fp);
			/* setting: Rho, MinTheta, MaxTheta, Thetasteps, MinTau, MaxTau, Tausteps, NumRuns */
			SimParDialog(infile, prob_fp);

#	ifdef USE_MPI
			PI_initslavequeue(PI_NumProcs);

			PI_SendParams(
				Freq,		/* SubsProDialog */
				kappa,		/* SubsProDialog */
				ksi,		/* SubsProDialog */
				NumCats,	/* SubsProDialog */
				GammaShape,	/* SubsProDialog */
				NumSamples,	/* DataDialog */
				NumBases,	/* DataDialog */
				NumVarPos,	/* DataDialog */
				MeanPWDist,	/* DataDialog */
/* change ;-) - done */
				/* Rho, */	/* SimParDialog */
				Rho_arr,	/* SimParDialog */
				Rhosteps,	/* SimParDialog */
				MinTau,		/* SimParDialog */
				MaxTau,		/* SimParDialog */
				Tausteps,	/* SimParDialog */
				MinTheta,	/* SimParDialog */
				MaxTheta,	/* SimParDialog */
				Thetasteps,	/* SimParDialog */
				NumRuns,	/* SimParDialog */
				tolerance,	/* SimParDialog */
				randseed,	/* random number seed */
				writeraw_optn);	/* write every simulation result? */
		} else { /* ! PI_IamMaster */
			PI_RecvParams(
				Freq,			/* SubsProDialog */
				&kappa,			/* SubsProDialog */
				&ksi,			/* SubsProDialog */
				&NumCats,		/* SubsProDialog */
				&GammaShape,		/* SubsProDialog */
				&NumSamples,		/* DataDialog */
				&NumBases,		/* DataDialog */
				&NumVarPos,		/* DataDialog */
				&MeanPWDist,		/* DataDialog */
/* change ;-) - done */
				/* &Rho, */		/* SimParDialog */
				Rho_arr,		/* SimParDialog */
				&Rhosteps,		/* SimParDialog */
				&MinTau,		/* SimParDialog */
				&MaxTau,		/* SimParDialog */
				&Tausteps,		/* SimParDialog */
				&MinTheta,		/* SimParDialog */
				&MaxTheta,		/* SimParDialog */
				&Thetasteps,		/* SimParDialog */
				&NumRuns,		/* SimParDialog */
				&tolerance,		/* SimParDialog */
				&randseed,		/* random number seed */
				&writeraw_optn);	/* write every simulation result? */
		} /* end if PI_IamMaster */
#	endif /* USE_MPI */

	/* gen = 20.0; */  /* never really used !!! delete? (HAS) */
	
	fprintf(stderr,"\n");

	time(&zeit1);

	/* mY = (double)NumBases/360.0/500.0; */ /* never used !!! delete? (HAS) */

	SetModel();

	DiscreteGamma(freqRate, CatRate, GammaShape, GammaShape, NumCats, 0);
	Categories=CAllocMem(NumBases*sizeof(short));

#	ifdef USE_OPENMP
		{
			int n;
			alltrees=(TTree  **)malloc(NumThreads * sizeof(TTree *));
			allnodes=(TNode ***)malloc(NumThreads * sizeof(TNode **));
			for (n=0;n<NumThreads; n++) {
				/* TO BE MADE THREADSAFE (HAS) */
				alltrees[n]=(TTree *)malloc(sizeof(TTree ));
			}
		}
#	else
		/* TO BE MADE THREADSAFE (HAS) */
		tree=(TTree *)malloc(sizeof(TTree));
		NumThreads = 1;
		MyThreadNum = 0;
#	endif
   
	/* init random seed */
	srand((unsigned)time(NULL));
#	ifdef FIXEDRAND
		randseed=-1;
		seedran(randseed);
		fprintf(stderr, "\nrandom number seed fixed to -1 for debugging!!!\n\n");
#	else
		randseed=-(long)rand();
		seedran(randseed);
#	endif

#	ifdef USE_MPI
		if (PI_IamMaster) {
#	endif /* USE_MPI */
			/* alloc and init 3D result array */
/* change ;-) - done */
			prob_com = f3tensor(0,Thetasteps-1, 0,Rhosteps-1, 0,Tausteps-1);
			for(f=0; f<Rhosteps; f++) {
				for(d=0; d<Thetasteps; d++) {
/* change ;-) - done */
					for(tt=0; tt<Tausteps; tt++) {
						prob_com[d][f][tt] = 0;
					}
				}
			}

			/* alloc vector for counting observed values */
			/* TO BE MADE THREADSAFE (HAS) */
			/* observed = dvector(0,1); */	/* not really used - delete? (HAS) */

			fprintf(prob_fp,"\n");
			fprintf(prob_fp,"The settings for this run were:\n");
			fprintf(prob_fp,"sample size:            %d\n", NumSamples);
			fprintf(prob_fp,"sequence length:        %d\n", NumBases);
			fprintf(prob_fp,"variable positions:     %d\n", NumVarPos);
			fprintf(prob_fp,"mean pairwise distance: %.2f\n", MeanPWDist);
			fprintf(prob_fp,"delta:                  %f\n", tolerance);
			fprintf(prob_fp,"simulations:            %ld\n", NumRuns);
/* change ;-) - done */
			fprintf(prob_fp,"\nRHO was set equal to\n\t");
			for (n=0; n<Rhosteps; n++)
				fprintf(prob_fp,"%.4g  ", Rho_arr[n]);
			fprintf(prob_fp,"\n\n");
			if (Rhosteps == 1)
				fprintf(prob_fp,"The matrix below contains appr. likelihood values,\n");
			else
				fprintf(prob_fp,"The matrices below contain appr. likelihood values,\n");

			fprintf(prob_fp,"where the lines correspond to THETA values of:\n\n");
			if ((MaxTheta == MinTheta) && (Thetasteps==1)) 
				fprintf(prob_fp,"%.4f  ", MinTheta);	/* only one Theta */
			else for(i=0;i<Thetasteps;i++)
				fprintf(prob_fp," %.4f ", MinTheta + ((MaxTheta-MinTheta)/(Thetasteps-1))*i);

			if(Tausteps == 0)
				fprintf(prob_fp,"\n(TAU undefined)");
			else {
				fprintf(prob_fp,"\n\nand the columns to TAU values of:\n\n");
				if ((MaxTau == MinTau) && (Tausteps==1)) 
					fprintf(prob_fp,"%.4f  ", MinTau);	/* only one Tau */
				else for(tt=0;tt<Tausteps;tt++)
					fprintf(prob_fp," %.4f ",MinTau + ((MaxTau-MinTau)/(Tausteps-1))*tt);
			}
#if 0
/* change ;-) - done (moved to Rho-loop */
			fprintf(prob_fp,"\n\n\nMatrix:\n");
#endif
			/* fclose(prob_fp); */
			fflush(prob_fp);

/* change ;-) - done */
/* for-loop RHO */
#if 0
			f=0;
			factor = 1.0/Rho;
#endif

			if (writeraw_optn) {
				if ( (rawfile = fopen(RAWFILE,"w")) == NULL) {
					fprintf(stderr,"ERROR: cannot open raw data file \"%s\"!", RAWFILE);
					exit(1);
				}
			}

			/**************/
			/* simulation */
			/**************/
			/* LATER MAYBE PARALLELIZE THE GRID (HAS) */
			/* if (PI_IamMaster) */
/* change ;-) - Rho-loop + Matrix head */
/* for-loop RHO */

			/* print Tau, Theta, Rho to params file */
			if ( (params_fp = fopen(PARAMSFILE,"w")) == NULL)
				fprintf(stderr,"params-file-problems");

			/***************************/
			fprintf(params_fp, "rho");

			for(f=0; f<Rhosteps; f++) {		/* start of "rho" loop */
				Rho = Rho_arr[f];
				/* f=0; */

				numlength=round(log10(Rho));
				if (numlength < 0.0) {
   					sprintf(FORMAT, "\t%%0.%df", (int)(-numlength));
				} else {
   					sprintf(FORMAT, "\t%%0.%df", 1);
				}
   				fprintf(params_fp, FORMAT, Rho);
			}
			fprintf(params_fp, "\n");

			/***************************/
			fprintf(params_fp, "theta");
			for(d=0; d<Thetasteps; d++) {		/* start of "theta" loop */

				/* compute current theta value in "theta" loop : */
				/*               Minimum  + step size                          * step number */
				if ((MaxTheta == MinTheta) && (Thetasteps==1))
					theta0 = (double)MinTheta;
				else
					theta0 = (double)(MinTheta + ((MaxTheta-MinTheta)/(Thetasteps-1)) * d);

				fprintf(params_fp,"\t%.6f", theta0);
			}
			fprintf(params_fp, "\n");

			/***************************/
			fprintf(params_fp, "tau");
			for(tt=0; tt<Tausteps; tt++) {	/* start of "Tau" loop */
/* change ;-) - check RHO==1 -> tau undef, done */

				Rho = Rho_arr[0];
				if((Rhosteps==1) && (0.99<Rho) && (Rho<1.01)) {
					Tau = 0.0;	/* Tau undefined, due to Rho = 1 */
				} else {
					/* compute current tau value in "Tau" loop : */
					if(MinTau == MaxTau) { /* no varying tau */
						/*   growth * Minimum      / (current theta / 2) */
						Tau = (double)MinTau;
					} else {
						/*   growth * (Minimum       + step size           * step number) / (current theta / 2) */
						Tau = ((double)(MinTau + (MaxTau-MinTau)/(Tausteps-1) * tt));
					}
				}
				fprintf(params_fp,"\t%.6f", Tau);
			}
			fprintf(params_fp, "\n");
			/***************************/
			fprintf(params_fp, "simulations");
			fprintf(params_fp, "\t%ld\n", NumRuns);

			fflush(matr_fp);

			/***************************/

			sprintf(SIMNUMFORMAT, "\t%%.%df", (int)round(log10(NumRuns)));
			/* fprintf(stderr, "XXX: %s\n", SIMNUMFORMAT); */

			for(f=0; f<Rhosteps; f++) {		/* start of "rho" loop */
				Rho = Rho_arr[f];
				/* f=0; */
				factor = 1.0/Rho;

				if((0.99<Rho) && (Rho<1.01))
					Taustepscount = 1;
				else	Taustepscount = Tausteps;

				numlength=round(log10(Rho));
				if (numlength < 0.0) {
   					sprintf(FORMAT, "%s_%%0.%df.%s", FILEPREFIX, (int)(-numlength), MATRIXFILEEXT);
				} else {
   					sprintf(FORMAT, "%s_%%0.%df.%s", FILEPREFIX, 1, MATRIXFILEEXT);
				}

   				sprintf(MATRIXFILE, FORMAT, Rho);
				fprintf(stdout, "Writing to: %s\n", MATRIXFILE);
   				/* fprintf(stdout, "rho=%f factor=%f len=%d %s\n", Rho, factor, (int)round(numlength), MATRIXFILE); */

				/* print Tau header of matrix file */
				if ( (matr_fp = fopen(MATRIXFILE,"w")) == NULL)
					fprintf(stderr,"matrix-file-problems");

				for(tt=0; tt<Taustepscount; tt++) {	/* start of "Tau" loop */
/* change ;-) - check RHO==1 -> tau undef, done */

					if((0.99<Rho) && (Rho<1.01)) {
						Tau = 0.0;	/* Tau undefined, due to Rho = 1 */
					} else {
						/* compute current tau value in "Tau" loop : */
						if(MinTau == MaxTau) { /* no varying tau */
							/*   growth * Minimum      / (current theta / 2) */
							Tau = (double)MinTau;
						} else {
							/*   growth * (Minimum       + step size           * step number) / (current theta / 2) */
							Tau = ((double)(MinTau + (MaxTau-MinTau)/(Tausteps-1) * tt));
						}
					}
					fprintf(matr_fp,"\t%.6f", Tau);
				}
				fprintf(matr_fp,"\n", Tau);

				/* print current Matrix head to "panel" file */
#if 0
				if ( (prob_fp = fopen(PANELFILE,"a")) == NULL)
					fprintf(stderr,"outfile-problems");
#endif

				if((0.99<Rho) && (Rho<1.01))
					fprintf(prob_fp,"\n\n\nMatrix (for Rho = %g), TAU undefined:\n", Rho);
				else
					fprintf(prob_fp,"\n\n\nMatrix (for Rho = %g):\n", Rho);

				/* fclose(prob_fp); */
				fflush(prob_fp);

				for(d=0; d<Thetasteps; d++) {		/* start of "theta" loop */
	
					/* compute current theta value in "theta" loop : */
					/*               Minimum  + step size                          * step number */
					if ((MaxTheta == MinTheta) && (Thetasteps==1))
						theta0 = (double)MinTheta;
					else
						theta0 = (double)(MinTheta + ((MaxTheta-MinTheta)/(Thetasteps-1)) * d);
	
					fprintf(matr_fp,"%.6f", theta0);
					fflush(matr_fp);

					/* tt=0; */	/* first tau step */
					for(tt=0; tt<Taustepscount; tt++) {	/* start of "Tau" loop */
/* change ;-) - check RHO==1 -> tau undef, done */
	
						if((0.99<Rho) && (Rho<1.01)) {
							Tt = 0.0;	/* Tau undefined, due to Rho = 1 */
						} else {
							/* compute current tau value in "Tau" loop : */
							if(MinTau == MaxTau) { /* no varying tau */
								/*   growth * Minimum      / (current theta / 2) */
								Tau = (double)MinTau;
								Tt = factor * ((double)Tau/ (theta0/2.0));
							} else {
								/*   growth * (Minimum       + step size           * step number) / (current theta / 2) */
								Tau = ((double)(MinTau + (MaxTau-MinTau)/(Tausteps-1) * tt));
								Tt = factor * ((double)Tau / (theta0/2.0));
								/* Tt = factor * (((double)(MinTau + (MaxTau-MinTau)/(Tausteps-1) * tt) / (theta0/2.0))); */
							}
						}

						if((0.99<Rho) && (Rho<1.01))
							printf("Simulating grid position: Rho=%f Theta=%f Tau=undef\n", 1.0, theta0);
						else	printf("Simulating grid position: Rho=%f Theta=%f Tau=%f\n", Rho, theta0, Tau);

						/* TODO: OMP_NUM_THREADS tree structures */
						/* TODO: result collector variables + reduction(+:) */
						/* pragma omp parallel private(num_positive,tree,Tlist) reduction(+:num_positive) */
						num_positive = 0;
	
#						ifdef USE_MPI
							{
								counter=0;
								num_positive = 0;
	
#if 0
#								ifdef USE_OPENMP
								{
									MyThreadNum = omp_get_thread_num();
									NumThreads  = omp_get_max_threads();
									tree  = alltrees[MyThreadNum];
									Tlist = allnodes[MyThreadNum];
								}
#								else
									/* TO BE MADE THREADSAFE (HAS) */
									tree=(TTree *)malloc(sizeof(TTree));
									NumThreads = 1;
									MyThreadNum = 0;
#								endif
#endif
								/* TO BE MADE THREADSAFE (HAS) */
									tree=(TTree *)malloc(sizeof(TTree));
									NumThreads = 1;
									MyThreadNum = 0;
   	
#								ifdef VERBOSE
									fprintf(stderr,"[%2d] distributing %lu simulations!!!\n", PI_Myid, NumRuns);
#								endif
	
/* change ;-) - no change necessary, due to single Rho */
								PI_SendDoSimulBlock(NumRuns,	/* total # of simulations todo */
	 									theta0,		/* theta0 */
	 									Tt,		/* Tt */
	 									Rho,		/* Rho */
	 									&num_positive);	/* number of hits received */
	
#								ifdef VERBOSE
									fprintf(stderr,"[%2d] sum of %lu positive of %lu simulation received!!!\n", PI_Myid, num_positive, NumRuns);
#								endif
	
							} /* end of OMP parallel part */
#						else /* ! USE_MPI */
#							pragma omp parallel private(MyThreadNum,NumThreads,tree,Tlist,tmp) reduction(+:num_positive)
							{
								counter=0;
								num_positive = 0;
	
#								ifdef USE_OPENMP
									{
										MyThreadNum = omp_get_thread_num();
										NumThreads  = omp_get_max_threads();
										tree  = alltrees[MyThreadNum];
										Tlist = allnodes[MyThreadNum];
									}
#								else /* ! USE_OPENMP */
									/* TO BE MADE THREADSAFE (HAS) */
									tree=(TTree *)malloc(sizeof(TTree));
									NumThreads = 1;
									MyThreadNum = 0;
#								endif /* ! USE_OPENMP */

#								ifdef VERBOSE
									if (MyThreadNum == 0) {
										if ((MaxTheta == MinTheta) && (Thetasteps==1)) 
											fprintf(stderr,"Theta=%.3f, ", MinTheta);
										else    fprintf(stderr,"Theta=%.3f, ", MinTheta + (MaxTheta-MinTheta)/(Thetasteps-1) * d);
										if ((MaxTheta == MinTheta) && (Thetasteps==1)) 
											fprintf(stderr,"Tau=%.3f, ", MinTau);
										else {
											if ((MaxTheta == MinTheta) && (Thetasteps==0)) 
												fprintf(stderr,"Tau=undefined\n\t");
											else	fprintf(stderr,"Tau=%.3f\n\t", MinTau + (MaxTau-MinTau)/(Tausteps-1) * tt);
										}
									}
#								endif /* VERBOSE */
	
/* change ;-) - no change neccessary, due to single Rho in loop */
								if (writeraw_optn) {
									if((0.99<Rho) && (Rho<1.01))
										fprintf(rawfile, "%f\t%f\tundef", 1.0, theta0);
									else	fprintf(rawfile, "%f\t%f\t%f", Rho, theta0, Tt);
									fflush(rawfile);
								}
	
#								pragma omp for 
								for(r=1;r<=NumRuns;r++) { /* FIRST PARALLELIZATION TARGET (HAS) */
#									pragma omp critical
									{ counter++; tmp = counter; }
#									ifdef VERBOSE
										/* output of dots */
										if(tmp%SIMPERDOT == 0) {
											fprintf(stderr,".");
											/* fprintf(stderr,"%d",MyThreadNum); */
											if(tmp%(10*SIMPERDOT) == 0) {
												fprintf(stderr," ");
												if(tmp%(60*SIMPERDOT) == 0) {
													fprintf(stderr,"\n\t");
												}
											}
										}
#									endif
									/* cunstruct random tree */
									/* grow_tree(factor,Tt); */		/* HAS TO BE THREAD-SAFE (HAS) */
									grow_tree(factor, Tt, &tree, &Tlist);
									/* simulate */
									seq_gen(tree,theta0,&num_positive, rawfile);	/* HAS TO BE THREAD-SAFE (HAS) */
									/* clean up */
									T1=tree->nodeList;
									while(T1 != NULL) {
										T2=T1;
										T1=T1->next;
										free(T2);
									}
									free(Tlist);
								}	/* end of "NumRuns" loop */
	
								if (writeraw_optn) { /* nextline */
									fprintf(rawfile, "\n");
									fflush(rawfile);
								}
	
							} /* end of OMP parallel part */
#						endif /* elif ! USE_MPI */

	
						prob_com[d][f][tt] = (double)num_positive; /* count in variable for global reduction(sum) (HAS) */
	
#						ifdef VERBOSE
#							ifdef USE_MPI
								fprintf(stderr,"[%2d] ", PI_Myid);
#							endif /* USE_MPI */
							if(counter%SIMPERDOT == 0) fprintf(stderr,".");
							fprintf(stderr," (%lu) ", num_positive);
							fprintf(stderr,"\n");
#						endif
	
						/* print current result to "panel" file */
#if 0
						if ( (prob_fp = fopen(PANELFILE,"a")) == NULL)
							fprintf(stderr,"outfile-problems");
#endif
/* change ;-) - no change necessary, output per column */
						fprintf(prob_fp," %5.4f", prob_com[d][f][tt]/NumRuns);
						/* fclose(prob_fp); */
						fflush(prob_fp);
	
						fprintf(matr_fp,SIMNUMFORMAT, (double)prob_com[d][f][tt]/(double)NumRuns);
						fflush(matr_fp);

						/* tt++; */
					}  /* end of "T" loop */
					/* while(tt < Tausteps); */
	
#if 0
					if ( (prob_fp = fopen(PANELFILE,"a")) == NULL)
						fprintf(stderr,"outfile-problems");
#endif
					fprintf(prob_fp,"\n");
					/* fclose(prob_fp); */
					fflush(prob_fp);
					fprintf(matr_fp,"\n");
					fflush(matr_fp);
	
				}  /* end of "theta" loop */
				fprintf(matr_fp,"\n");
			}  /* end of "rho" loop */
#	ifdef USE_MPI
		} else { /* slave */
			slave_loop(rawfile);
		}
#	endif /* USE_MPI */

#	ifdef USE_MPI
		if (PI_IamMaster) {
#	endif /* USE_MPI */
			/* fprintf(stderr,"\n"); */

			if (writeraw_optn) fclose(rawfile);

#if 0
			if ( (prob_fp = fopen(PANELFILE,"a")) == NULL) {
  				fprintf(stderr,"ERROR: cannot open outputfile \"%s\"!", PANELFILE);
  				exit(1);
			}
#endif


#	ifdef USE_MPI
#			ifdef PVERBOSE
				fprintf(stderr,"[%2d] sending done2\n", PI_Myid);
#			endif /* PVERBOSE */
			PI_SendDone();
			MPI_Finalize();

#if 0
			if ( (prob_fp = fopen(PANELFILE,"a")) == NULL)
				fprintf(stderr,"outfile-problems");
#endif
			time(&zeit2);
			fprintf(prob_fp,"\nFinished! Time needed: %.4f (min)\n", difftime(zeit2, zeit1)/60.);
			fprintf(stdout,"\nFinished! Time needed: %.4f (min)\n", difftime(zeit2, zeit1)/60.);
			fclose(prob_fp);
#if 0
	fprintf(stderr,"[%2d] dying... master\n", PI_Myid);
#	endif /* 0 */
		} else {
#if 0
	fprintf(stderr,"[%2d] dying...\n", PI_Myid);
#	endif /* 0 */
		}
#	else /* ! USE_MPI */
		time(&zeit2);
#if 0
		if ( (prob_fp = fopen(PANELFILE,"a")) == NULL)
			fprintf(stderr,"outfile-problems");
#endif
		fprintf(prob_fp,"\nFinished! Time needed: %.4f (min)\n", difftime(zeit2, zeit1)/60.);
		fprintf(stdout,"\nFinished! Time needed: %.4f (min)\n", difftime(zeit2, zeit1)/60.);
		fclose(prob_fp);
#	endif /* ! USE_MPI */

				fclose(matr_fp);
	return 0;
}

