/*
 * mpiiphula.c
 *
 *
 * Part of (p)IPHULA 1.16 (May 2007)
 *
 * (c) 2004-2007 by Heiko A. Schmidt, Jutta Buschbom, 
 *                  Gunter Weiss, and Arndt von Haeseler
 * (c) 1998-2004 by Gunter Weiss and Arndt von Haeseler
 *
 * All parts of the source except where indicated are distributed 
 * under the GNU public licence 2. 
 * See http://www.opensource.org/licenses/gpl-license.php details.
 *
 * ($Id: mpiiphula.c,v 1.3 2005/09/30 11:01:03 hschmidt Exp $)
 *
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "mpiiphula.h"

#if 0
# define EXTERN extern
#endif


#define PVERBOSE_SEND_START 1
#define PVERBOSE_SEND_END   1
#define PVERBOSE_RECV_START 1
#define PVERBOSE_RECV_END   1
#define VERBOSE             1

#undef PVERBOSE_SEND_START
#undef PVERBOSE_SEND_END
#undef PVERBOSE_RECV_START
#undef PVERBOSE_RECV_END
#undef VERBOSE
#if 0
#endif
 
int PI_IamMaster;
int PI_IamSlave;
int PI_Myid;
int PI_MyMaster;
int PI_NumProcs;
MPI_Comm PI_Comm;

uli *simulsent;
uli *simulrecved;
uli *dosimulsent;
uli *dosimulrecved;

uli PI_simulsent = 0;          /* # of  */
uli PI_simulsentn = 0;          /* # of  */
uli PI_simulrecved = 0;        /* # of  */

/*********************************************************************
*  debugging                                                         *
*********************************************************************/

static int deb_num = 0;
void fprintfdebug(FILE *fp, char s[], char f[], int l)
{
        fprintf(fp, "[%2d] %s (%s:%d)\n", deb_num++, s, f, l);
}
/*
fprintfdebug(stderr, "debug message", __FILE__, __LINE__);
*/


/*********************************************************************
*  miscellaneous utilities                                           *
*********************************************************************/

/*********************************************************************
*********************************************************************/
void mi_maerror(char *message)
{
        fprintf(stdout, "\n\n\nUnable to proceed (lack of memory: %s)\n\n", message);
#       if USE_MPI
                PI_Finalize();
#       endif
        exit(1);
} /* mi_maerror */

/******************/

int *mi_new_ivector(int n)
{
        int *v;

        v = (int *) calloc((size_t) n, sizeof(int));
        if (v == NULL) mi_maerror("step 1 in new_ivector");

        return v;
} /* mi_new_ivector */

/******************/

void mi_free_ivector(int *v)
{
        free((int *) v);
} /* mi_free_ivector */


/******************/
uli *mi_new_ulivector(int n)
{
        uli *v;

        v = (uli *) calloc((size_t) n, sizeof(uli));
        if (v == NULL) mi_maerror("step 1 in new_ulivector");

        return v;
} /* mi_new_ulivector */

/******************/

void mi_free_ulivector(uli *v)
{
        free((uli *) v);
} /* mi_free_ulivector */


/******************/
 
void PI_Printerror(FILE *of, int id, int err)
{
        char  errstr[MPI_MAX_ERROR_STRING];
        int   errstrlen;

  if ((err > MPI_SUCCESS) && (err <= MPI_ERR_LASTCODE)) {
    MPI_Error_string(err, errstr, &errstrlen);
    fprintf(of, "[%2d] MPI ERROR %d : %s\n", id, err, errstr);
    }
  else {
    if (err == MPI_SUCCESS) 
      fprintf(of, "[%2d] MPI ERROR %d : No error\n", id, err);
    else
      fprintf(of, "[%2d] MPI ERROR %d : unknown error number\n", id, err);
  }
} /* PI_Printerror */

/******************/

/*********************************************************************
*  sending/receiving the important sizes and parameter  (M->S)       *
*********************************************************************/
 
void PI_SendParams(
	double *Freq,		/* SubsProDialog */
	double  kappa,		/* SubsProDialog */
	double  ksi,		/* SubsProDialog */
	int     NumCats,		/* SubsProDialog */
	double  GammaShape,	/* SubsProDialog */
	int     NumSamples,	/* DataDialog */
	int     NumBases,	/* DataDialog */
	int     NumVarPos,	/* DataDialog */
	double  MeanPWDist,	/* DataDialog */
/* change ;-) - done */
	/* double Rho, */	/* SimParDialog */
	double *Rho_arr,		/* SimParDialog */
	int     Rhosteps,	/* SimParDialog */
	double  MinTau,		/* SimParDialog */
	double  MaxTau,		/* SimParDialog */
	int     Tausteps,	/* SimParDialog */
	double  MinTheta,	/* SimParDialog */
	double  MaxTheta,	/* SimParDialog */
	int     Thetasteps,	/* SimParDialog */
	long    NumRuns,		/* SimParDialog */
	double  tolerance,	/* SimParDialog */
	long    randseed,	/* random number seed */
	int     writeraw_optn)	/* write every simulation result? */
{
#   define NUMDBL 13
#   define NUMINT  8
#   define NUMLINT 2
	int    ints[NUMINT];
	uli    lints[NUMLINT];
	double doubles[NUMDBL + MAX_RHO_STEPS];
	MPI_Datatype Dtypes[3] =    {MPI_INT, MPI_LONG_INT, MPI_DOUBLE};
	int          Dtypelens[3] = {NUMINT, NUMLINT, NUMDBL + MAX_RHO_STEPS};
	MPI_Aint     Dtypeaddr[3];
	MPI_Datatype PI_Params;
	int          dest;
	int          n;		/* counter variable */
	int          error;

	doubles[0]  = Freq[0];		/* SubsProDialog */
	doubles[1]  = Freq[1];		/* SubsProDialog */
	doubles[2]  = Freq[2];		/* SubsProDialog */
	doubles[3]  = Freq[3];		/* SubsProDialog */
	doubles[4]  = kappa;		/* SubsProDialog */
	doubles[5]  = ksi;		/* SubsProDialog */
	doubles[6]  = GammaShape;	/* SubsProDialog */
	doubles[7]  = MeanPWDist;	/* DataDialog */
	doubles[8]  = MinTau;		/* SimParDialog */
	doubles[9] = MaxTau;		/* SimParDialog */
	doubles[10] = MinTheta;		/* SimParDialog */
	doubles[11] = MaxTheta;		/* SimParDialog */
	doubles[12] = tolerance;	/* SimParDialog */
/* change ;-) - done */
	for (n=0; n<MAX_RHO_STEPS; n++) {
		doubles[NUMDBL + n]  = Rho_arr[n];	/* SimParDialog */
	}
	ints[0]     = NumCats;		/* SubsProDialog */
	ints[1]     = NumSamples;	/* DataDialog */
	ints[2]     = NumBases;		/* DataDialog */
	ints[3]     = NumVarPos;	/* DataDialog */
	ints[4]     = Tausteps;		/* SimParDialog */
	ints[5]     = Thetasteps;	/* SimParDialog */
	ints[6]     = Rhosteps;		/* SimParDialog */
	ints[7]     = writeraw_optn;	/* write every simulation result? */
	lints[0]    = NumRuns;		/* SimParDialog */
	lints[1]    = randseed;		/* random number seed */

	MPI_Address(ints,     Dtypeaddr);
	MPI_Address(lints,   (Dtypeaddr+1));
	MPI_Address(doubles, (Dtypeaddr+2));

	MPI_Type_struct(3, Dtypelens, Dtypeaddr, Dtypes, &PI_Params);
	MPI_Type_commit(&PI_Params);

	for (dest=1; dest<PI_NumProcs; dest++) {

#		ifdef PVERBOSE_SEND_START
			fprintf(stderr, "[%2d] -> (%2d) Sent Params (TAG=%d)\n", PI_Myid, dest, PI_PARAMS);
#		endif /* PVERBOSE_SEND_START */

		error = MPI_Ssend(MPI_BOTTOM, 1, PI_Params, dest, PI_PARAMS, PI_Comm);
		if (error != MPI_SUCCESS)
			PI_Printerror(stderr, 600+PI_Myid, error);

#		ifdef PVERBOSE_SEND_END
			fprintf(stderr, "[%2d] ... (%2d) Sent Params\n", PI_Myid, dest);
#		endif /* PVERBOSE_SEND_END */

	} /* for each slave */

	MPI_Type_free(&PI_Params);

#	ifdef PVERBOSE_SEND_END
		fprintf(stderr, "[%2d] ... Sent Params\n", PI_Myid);
#	endif /* PVERBOSE_SEND_END */

#	undef NUMINT
#	undef NUMLINT
#	undef NUMDBL
} /* PI_SendParams */


/******************/

void PI_RecvParams(
	double  *Freq,		/* SubsProDialog */
	double  *kappa,		/* SubsProDialog */
	double  *ksi,		/* SubsProDialog */
	int     *NumCats,	/* SubsProDialog */
	double  *GammaShape,	/* SubsProDialog */
	int     *NumSamples,	/* DataDialog */
	int     *NumBases,	/* DataDialog */
	int     *NumVarPos,	/* DataDialog */
	double  *MeanPWDist,	/* DataDialog */
/* change ;-) - done */
	double  *Rho_arr,	/* SimParDialog */
	int     *Rhosteps,	/* SimParDialog */
	double  *MinTau,	/* SimParDialog */
	double  *MaxTau,	/* SimParDialog */
	int     *Tausteps,	/* SimParDialog */
	double  *MinTheta,	/* SimParDialog */
	double  *MaxTheta,	/* SimParDialog */
	int     *Thetasteps,	/* SimParDialog */
	long    *NumRuns,	/* SimParDialog */
	double  *tolerance,	/* SimParDialog */
	long    *randseed,	/* random number seed */
	int     *writeraw_optn)	/* write every simulation result? */
{
#	define NUMDBL 13
#	define NUMINT  8
#	define NUMLINT 2

	int    ints[NUMINT];
	uli    lints[NUMLINT];
	double doubles[NUMDBL + MAX_RHO_STEPS];
	MPI_Datatype Dtypes[3] =    {MPI_INT, MPI_LONG_INT, MPI_DOUBLE};
	int          Dtypelens[3] = {NUMINT, NUMLINT, NUMDBL + MAX_RHO_STEPS};
	MPI_Aint     Dtypeaddr[3];
	MPI_Datatype PI_Params;
	MPI_Status   stat;
	int          n;		/* counter variable */
	int          error;
 
#	ifdef VERBOSE
		fprintf(stderr, "[%2d] entering RecvParams\n", PI_Myid);
#	endif /* VERBOSE */
	MPI_Address(ints,     Dtypeaddr);
	MPI_Address(lints,   (Dtypeaddr+1));
	MPI_Address(doubles, (Dtypeaddr+2));

	MPI_Type_struct(3, Dtypelens, Dtypeaddr, Dtypes, &PI_Params);
	MPI_Type_commit(&PI_Params);

#	ifdef PVERBOSE_RECV_START
		fprintf(stderr, "[%2d] Probing for parameters\n", PI_Myid);
#	endif /* PVERBOSE_RECV_START */

	error = MPI_Probe(PI_MyMaster, MPI_ANY_TAG, PI_Comm, &stat);

#	ifdef PVERBOSE_RECV_END
		fprintf(stderr, "[%2d] X... received TAG=%d=%d\n", PI_Myid, stat.MPI_TAG, PI_PARAMS);
#	endif /* PVERBOSE_RECV_END */

	if (error != MPI_SUCCESS)
		PI_Printerror(stderr, 700+PI_Myid, error);
	if (stat.MPI_TAG != PI_PARAMS) {
		if (stat.MPI_TAG == PI_DONE) {
			PI_RecvDone();
			MPI_Finalize();
			exit(1);
		} else {
			fprintf(stderr, "[%2d]a Error: unexpected TAG (%d) received...\n", PI_Myid, stat.MPI_TAG);
			MPI_Finalize();
			exit(1);
		}
	}

	/* error = MPI_Recv(MPI_BOTTOM, 1, PI_Params, PI_MyMaster, PI_PARAMS, PI_Comm, &stat); */
	error = MPI_Recv(MPI_BOTTOM, 1, PI_Params, PI_MyMaster, PI_PARAMS, PI_Comm, &stat);
	/* error = MPI_Recv(MPI_BOTTOM, 1, PI_Params, PI_MyMaster, MPI_ANY_TAG, PI_Comm, &stat); */
	if (error != MPI_SUCCESS)
		PI_Printerror(stderr, 700+PI_Myid, error);
	if (stat.MPI_TAG != PI_PARAMS) {
		fprintf(stderr, "[%2d]b Error: unexpected TAG (%d) received...\n", PI_Myid, (int)stat.MPI_TAG);
#if 0
		MPI_Finalize();
		exit(1);
#endif
	}

*tolerance=-0.2;
/* fprintf(stderr, "[%2d]2 tol: %f\n", PI_Myid, *tolerance); */
	Freq[0]        = doubles[0];	/* SubsProDialog */
	Freq[1]        = doubles[1];	/* SubsProDialog */
	Freq[2]        = doubles[2];	/* SubsProDialog */
	Freq[3]        = doubles[3];	/* SubsProDialog */
	*kappa         = doubles[4];	/* SubsProDialog */
	*ksi           = doubles[5];	/* SubsProDialog */
	*GammaShape    = doubles[6];	/* SubsProDialog */
	*MeanPWDist    = doubles[7];	/* DataDialog */
/* change ;-) - done */
	*MinTau        = doubles[8];	/* SimParDialog */
	*MaxTau        = doubles[9];	/* SimParDialog */
	*MinTheta      = doubles[10];	/* SimParDialog */
	*MaxTheta      = doubles[11];	/* SimParDialog */
	*tolerance     = doubles[12];	/* SimParDialog */
	for (n=0; n<MAX_RHO_STEPS; n++) {
		doubles[NUMDBL + n]  = Rho_arr[n];	/* SimParDialog */
	}

	*NumCats       = ints[0];	/* SubsProDialog */
	*NumSamples    = ints[1];	/* DataDialog */
	*NumBases      = ints[2];	/* DataDialog */
	*NumVarPos     = ints[3];	/* DataDialog */
	*Tausteps      = ints[4];	/* SimParDialog */
	*Thetasteps    = ints[5];	/* SimParDialog */
/* change ;-) - done */
	*Rhosteps      = ints[6];	/* SimParDialog */
	*writeraw_optn = ints[7];	/* write every simulation result? */

	*NumRuns       = lints[0];	/* SimParDialog */
	*randseed      = lints[1];	/* random number seed */
 
/* fprintf(stderr, "[%2d]2 tol: %f\n", PI_Myid, *tolerance); */
	MPI_Type_free(&PI_Params);
 
#	undef NUMINT
#	undef NUMLINT
#	undef NUMDBL
} /* PI_RecvParams */
 

/*********************************************************************/
/*********************************************************************/
/*********************************************************************/
/*********************************************************************/
/*********************************************************************/
/*********************************************************************/
/*********************************************************************/
/*********************************************************************/
/*********************************************************************/
/*********************************************************************/
/*********************************************************************/
/*********************************************************************/
/*********************************************************************
*  procedures to send the splits of a puzzle tree to the master      *
*********************************************************************/

void PI_SendSimulBlock(uli rcv_pos, 
		       uli rcv_simul)
{
#	define NUMULI 2
	int               error;
	uli               ulis[NUMULI];

	ulis[0] = rcv_pos;
	ulis[1] = rcv_simul;

#	ifdef PVERBOSE_SEND_START
		fprintf(stderr, "[%2d] sending back results (TAG=%d) positive=%ld sim=%ld\n", PI_Myid, PI_SIMULRESULTS, rcv_pos, rcv_simul);
#	endif /* PVERBOSE_SEND_START */

	error = MPI_Ssend(ulis, NUMULI, MPI_UNSIGNED_LONG, PI_MyMaster, PI_SIMULRESULTS, PI_Comm);
	if (error != MPI_SUCCESS)
		PI_Printerror(stderr, 1800+PI_Myid, error);

#	ifdef PVERBOSE_SEND_END
		fprintf(stderr, "[%2d] ... sent\n", PI_Myid);
#	endif /* PVERBOSE_SEND_END */

 
#	undef NUMULI
} /* PI_SendSimulBlock */
 
/******************/

void PI_RecvSimulBlock(uli *rcv_pos, 
		       uli *rcv_simul)
{
#	define NUMULI 2
	MPI_Status         stat;
	int                error;
	int                dest;
	uli                ulis[NUMULI];

#	ifdef PVERBOSE_RECV_START
		fprintf(stderr, "[%2d] receiving SimulBlock results\n", PI_Myid);
#	endif /* PVERBOSE_RECV_START */

	error = MPI_Recv(ulis, NUMULI, MPI_UNSIGNED_LONG, MPI_ANY_SOURCE, PI_SIMULRESULTS, PI_Comm, &stat);
	if (error != MPI_SUCCESS)
		PI_Printerror(stderr, 1900+PI_Myid, error);

	dest       = stat.MPI_SOURCE; 
	*rcv_pos   = ulis[0];
	*rcv_simul = ulis[1];
 
#	ifdef PVERBOSE_RECV_END
		fprintf(stderr, "[%2d] ... received %lu positive hits of %lu <- (%2d, TAG=%d=%d)\n", PI_Myid, *rcv_pos, *rcv_simul, dest, stat.MPI_TAG, PI_SIMULRESULTS);
#	endif /* PVERBOSE_RECV_END */

	PI_putslave(dest); 
 
#	undef NUMULI
} /* PI_RecvSimulBlock */
 
 
/*********************************************************************
*  procedures to send the request to compute puzzle tree             *
*********************************************************************/

/* send all <puzzlings> requests and receive results */
void PI_SendDoSimulBlock(uli    simulations, /* total # of simulations todo */
			 double theta,       /* theta0 */
			 double tau,         /* Tt */
/* change ;-) - no change necessary, due to single Rho */
			 double rho,         /* Rho */
			 uli   *positive)    /* number of hits received */
{
#	define NUMULI 1
#	define NUMDBL 3
	uli           ulongs[NUMULI];
	double        doubles[NUMDBL];

	MPI_Datatype Dtypes[2] =    {MPI_UNSIGNED_LONG, MPI_DOUBLE};
	int          Dtypelens[2] = {NUMULI, NUMDBL};
	MPI_Aint     Dtypeaddr[2];
	MPI_Datatype PI_Simul;

	int           dest;               /* destination process */
	int           error;              /* MPI error code      */
	uli           numsimul;           /* simaltions scheduled for 1 chunk */
	uli           dotsprinted   = 0;  /* simaltions we are waiting for */
	uli           tmp;
	uli           simulsent     = 0;  /* simaltions we are waiting for */
	uli           simulreceived = 0;  /* simaltions we received back */
	schedtype     sched;              /* scheduler data */
	uli           rcv_pos;            /* number of positives */
	uli           rcv_simul;          /* number of received simulations */

#	ifdef VERBOSE
		fprintf(stderr, "[%2d] entering PI_SendDoSimulBlock\n", PI_Myid);
#	endif /* VERBOSE */
	doubles[0] = theta; /* theta0 */
	doubles[1] = tau;   /* Tt */
/* change ;-) - no change necessary, due to single Rho */
	doubles[2] = rho;   /* Rho */

	MPI_Address(ulongs,     Dtypeaddr);
	MPI_Address(doubles, (Dtypeaddr+1));

	MPI_Type_struct(2, Dtypelens, Dtypeaddr, Dtypes, &PI_Simul);
	MPI_Type_commit(&PI_Simul);

	initsched(&sched, simulations, PI_NumProcs-1, 4);
	/* numsimul = sc(&sched); */
	numsimul = SCHEDALG_PUZZLING(&sched);
	ulongs[0] = numsimul;

/* change ;-) - no change necessary, due to single Rho */
#	ifdef PVERBOSE
		fprintf(stderr, "\nSimulating: %lu (theta=%.3f, 'tau'=%.3f, rho=%.3f)\n\t", simulations, theta, tau, rho);
#	endif /* PVERBOSE */
	while (numsimul > 0) {
		if (PI_emptyslave()) {
#			ifdef PVERBOSE_RECV_START
				fprintf(stderr, "[%2d] waiting for results\n", PI_Myid);
#			endif /* PVERBOSE_RECV_START */
			PI_RecvSimulBlock(&rcv_pos, &rcv_simul);
			simulsent     -= rcv_simul;
			simulreceived += rcv_simul;
			*positive     += rcv_pos;

#			ifdef PVERBOSE
				/* output of dots */
				tmp = (int)floor(simulreceived / SIMPERDOT);
				while (dotsprinted < tmp) {
					dotsprinted++;
					if(dotsprinted == 1) fprintf(stderr,"\t");
					fprintf(stderr,".");
					/* fprintf(stderr,"%lu/%d/%d ", simulreceived, dotsprinted, tmp); */
					if((dotsprinted%10) == 0) {
						fprintf(stderr," ");
						if((dotsprinted%50) == 0) {
							fprintf(stderr,"\n\t");
						}
					}
				}
				fflush(stderr);
#			endif /* PVERBOSE */

#			ifdef PVERBOSE_RECV_END
				fprintf(stderr, "[%2d] ... received %lu positive hits of %lu\n", PI_Myid, rcv_pos, rcv_simul);
#			endif /* PVERBOSE_RECV_END */
		}

		dest = PI_getslave();

#		ifdef PVERBOSE_SEND_START
			fprintf(stderr, "[%2d] Sending %lu DOSimuls (TAG: %d) Signal -> %d\n", PI_Myid, numsimul, PI_DOSIMUL, dest);
#		endif /* PVERBOSE_SEND_START */

		error = MPI_Ssend(MPI_BOTTOM, 1, PI_Simul, dest, PI_DOSIMUL, PI_Comm);
		if (error != MPI_SUCCESS)
			PI_Printerror(stderr, 2100+PI_Myid, error);

#		ifdef PVERBOSE_SEND_END
			fprintf(stderr, "[%2d] ... Sent DOSimuls Signal -> %d\n", PI_Myid, dest);
#		endif /* PVERBOSE_SEND_END */

		simulsent    += numsimul;
		PI_simulsent += numsimul;
		PI_simulsentn ++;

		numsimul = SCHEDALG_PUZZLING(&sched);
		ulongs[0] = numsimul;

	} /* while something left to send */

	/* while waiting for results */
	while (simulsent > 0) {
#		ifdef PVERBOSE_RECV_START
			fprintf(stderr, "[%2d] waiting for results\n", PI_Myid);
#		endif /* PVERBOSE_RECV_START */
		/* receive next result */
		PI_RecvSimulBlock(&rcv_pos, &rcv_simul);
		/* PP_RecvSplitsBlock(&tx, &bs, &bp, &psteptreelist, &psteptreenum, &psteptreesum); */
		simulsent     -= rcv_simul;
		simulreceived += rcv_simul;
		*positive     += rcv_pos;

#		ifdef PVERBOSE
			/* output of dots */
			tmp = (int)floor(simulreceived / SIMPERDOT);
			while (dotsprinted < tmp) {
				dotsprinted++;
				if(dotsprinted == 1) fprintf(stderr,"\t");
				fprintf(stderr,".");
				/* fprintf(stderr,"%lu/%d/%d ", simulreceived, dotsprinted, tmp); */
				if((dotsprinted%10) == 0) {
					fprintf(stderr," ");
					if((dotsprinted%50) == 0) {
						fprintf(stderr,"\n\t");
					}
				}
			}
			fflush(stderr);
#		endif /* PVERBOSE */

#		ifdef PVERBOSE_RECV_END
			fprintf(stderr, "[%2d] ... received %lu positive hits of %lu\n", PI_Myid, rcv_pos, rcv_simul);
#		endif /* PVERBOSE_RECV_END */
	}

	MPI_Type_free(&PI_Simul);
 
#	undef NUMULI
#	undef NUMDBL
} /* PI_SendDoSimulBlock */

#if 0
#		ifdef VERBOSE_TEST
			/* output of dots */
			if(tmp%SIMPERDOT == 0) {
				fprintf(stderr,".");
				/* fprintf(stderr,"%d",MyThreadNum); */
				if(tmp%(10*SIMPERDOT) == 0) {
					fprintf(stderr," ");
					if(tmp%(60*SIMPERDOT) == 0) {
						fprintf(stderr,"\n\t");
					}
				}
			}
#		endif
#endif
/******************/


void PI_RecvDoSimulBlock(uli *simulations, /* # of simulations todo */
			 double *theta,    /* theta0 */
			 double *tau,      /* Tt */
/* change ;-) - no change necessary, due to single Rho */
			 double *rho)      /* Rho */
{
#	define NUMULI 1
#	define NUMDBL 3
	uli           ulongs[NUMULI];
	double        doubles[NUMDBL];

	MPI_Datatype Dtypes[2] =    {MPI_UNSIGNED_LONG, MPI_DOUBLE};
	int          Dtypelens[2] = {NUMULI, NUMDBL};
	MPI_Aint     Dtypeaddr[2];
	MPI_Datatype PI_Simul;

	MPI_Status    stat;
	int           error;
 
	MPI_Address(ulongs,     Dtypeaddr);
	MPI_Address(doubles, (Dtypeaddr+1));

	MPI_Type_struct(2, Dtypelens, Dtypeaddr, Dtypes, &PI_Simul);
	MPI_Type_commit(&PI_Simul);

#	ifdef PVERBOSE_RECV_START
		fprintf(stderr, "[%2d] Receiving: DOSimulBlock Signal\n", PI_Myid);
#	endif /* PVERBOSE_RECV_START */
 
	error = MPI_Recv(MPI_BOTTOM, 1, PI_Simul, PI_MyMaster, PI_DOSIMUL, PI_Comm, &stat);
	if (error != MPI_SUCCESS)
		PI_Printerror(stderr, 2100+PI_Myid, error);

	*theta       = doubles[0]; /* theta0 */
	*tau         = doubles[1];   /* Tt */
/* change ;-) - no change necessary, due to single Rho */
	*rho         = doubles[2];   /* Rho */
	*simulations = ulongs[0];


#	ifdef PVERBOSE_RECV_END
/* change ;-) - no change necessary, due to single Rho */
		fprintf(stderr, "[%2d] getting %lu DOSimulBlocks (TAG=%d=%d, theta=%f, tau=%f, rho=%f)\n", PI_Myid, *simulations, stat.MPI_TAG, PI_DOSIMUL, *theta, *tau, *rho);
#	endif /* PVERBOSE_RECV_END */
 
	MPI_Type_free(&PI_Simul);
 
#	undef NUMULI
#	undef NUMDBL
} /* PI_RecvDoPermutBlock */

/*********************************************************************
*  procedures to make the slaves to finalize                         *
*********************************************************************/

void PI_SendDone()
{
	int           dest;
	int           error;
 
#	ifdef PVERBOSE_SEND_START
		fprintf(stderr, "[%2d] Sending: DONE Signal\n", PI_Myid);
#	endif /* PVERBOSE_SEND_START */
 
	for (dest=1; dest<PI_NumProcs; dest++) {
 
		error = MPI_Ssend(&dest, 0, MPI_INT, dest, PI_DONE, PI_Comm);
		if (error != MPI_SUCCESS)
			PI_Printerror(stderr, 2100+PI_Myid, error);
 
	} /* for each slave */
 
#	ifdef PVERBOSE_SEND_END
		fprintf(stderr, "[%2d] ... Sent DONE Signal (TAG=%d)\n", PI_Myid, PI_DONE);
#	endif /* PVERBOSE_SEND_END */
 
} /* PI_SendDone */

/******************/



void PI_RecvDone()
{
	int           dummy;
	MPI_Status    stat;
	int           error;
 
#	ifdef PVERBOSE_RECV_START
		fprintf(stderr, "[%2d] Receiving: DONE Signal\n", PI_Myid);
#	endif /* PVERBOSE_RECV_START */
 
	error = MPI_Recv(&dummy, 0, MPI_INT, PI_MyMaster, PI_DONE, PI_Comm, &stat);

	if (error != MPI_SUCCESS)
		PI_Printerror(stderr, 2100+PI_Myid, error);

#	ifdef PVERBOSE_RECV_END
		fprintf(stderr, "[%2d] ... DONE Signal (TAG=%d=%d)\n", PI_Myid, PI_DONE, stat.MPI_TAG);
#	endif /* PVERBOSE_RECV_END */
 
} /* PI_RecvDone */

/*********************************************************************
*  procedures to initialize and cleanup                              *
*********************************************************************/

void PI_Init(int *argc, char **argv[])
{
	MPI_Init(argc, argv);
	PI_Comm = MPI_COMM_WORLD;
	MPI_Comm_rank(PI_Comm, &PI_Myid);
	MPI_Comm_size(PI_Comm, &PI_NumProcs);

	if (PI_NumProcs <= 1) {
		fprintf(stderr, "\nHalt: pIPHULA needs at least 2 processes for a parallel run!!!\n\n");
		MPI_Finalize();
		exit(1);
	}

	PI_MyMaster = 0;

	if (PI_Myid == PI_MyMaster) {
		PI_IamMaster=1;
		PI_IamSlave =0;
		PI_initslavequeue(PI_NumProcs - 1);

		simulsent     = mi_new_ulivector(PI_NumProcs);
		simulrecved   = mi_new_ulivector(PI_NumProcs);
		dosimulsent   = mi_new_ulivector(PI_NumProcs);
		dosimulrecved = mi_new_ulivector(PI_NumProcs);

		{
			int n;
			for (n = 0; n<PI_NumProcs; n++){
				simulsent     [n] = -1; 
				simulrecved   [n] = -1; 
				dosimulsent   [n] = -1; 
				dosimulrecved [n] = -1; 
			}
		}
	} else {
		PI_IamMaster=0;
		PI_IamSlave =1;
	}

} /* PI_Init */

/******************/

void PI_Finalize()
{
	if (PI_IamMaster) {
		mi_free_ivector(freeslaves);

		mi_free_ulivector(simulsent);
		mi_free_ulivector(simulrecved);
		mi_free_ulivector(dosimulsent);
		mi_free_ulivector(dosimulrecved);
	}

	MPI_Finalize();
} /* PI_Finalize */

/***********************************************************
*  main part of the slave process                          *
***********************************************************/

#if 0
int slave_loop(int argc, char *argv[])
{
	int        notdone;
	MPI_Status stat;
	uli        simul;

	/* PI_RecvParams(&Maxspc, &Maxsite, &numcats, &Numptrn, &tpmradix, &outgroup, &fracconst, &randseed, &fixedorder_optn, &consensus_optn); */
	PI_RecvParams(
		Freq,		/* SubsProDialog */
		&kappa,		/* SubsProDialog */
		&ksi,		/* SubsProDialog */
		&NumCats,	/* SubsProDialog */
		&GammaShape,	/* SubsProDialog */
		&NumSamples,	/* DataDialog */
		&NumBases,	/* DataDialog */
		&NumVarPos,	/* DataDialog */
		&MeanPWDist,	/* DataDialog */
/* change ;-) - no change necessary - if 0 */
		&Rho,		/* SimParDialog */
		&MinTau,	/* SimParDialog */
		&MaxTau,	/* SimParDialog */
		&Tausteps,	/* SimParDialog */
		&MinTheta,	/* SimParDialog */
		&MaxTheta,	/* SimParDialog */
		&Thetasteps,	/* SimParDialog */
		&NumRuns,	/* SimParDialog */
		&tolerance,	/* SimParDialog */
		&randseed);	/* random number seed */

	initrandom(randseed);

	MPI_Probe(PI_MyMaster, MPI_ANY_TAG, PI_Comm, &stat);

	notdone = (stat.MPI_TAG != PI_DONE);
	if (!notdone) {
		PI_RecvDone();
	}

	while (notdone) {
		switch (stat.MPI_TAG) {
			case PI_DOSIMUL: {
				uli r;
				/* Receive specs for simulation */
/* change ;-) - no change necessary - if 0 */
				PI_RecvDoSimulBlock(&simul, &theta0, &Tt, &Rho);
				for (r = 3; r < simul; r++) {
					/* do something */
				}

				/* send results */
				PI_SendSimulBlock(pos, simul);
				break;
				}
			default: {
				fprintf(stderr, "ERROR: Unknown Message Tag received\n");
				MPI_Abort(PI_Comm, 1);
				}
		} /* switch stat.MPI_TAG */ 

		MPI_Probe(PI_MyMaster, MPI_ANY_TAG, PI_Comm, &stat);
		notdone = (stat.MPI_TAG != PI_DONE);
		if (!notdone)
			PI_RecvDone();

	} /* while notdone */

	return	(0);
} /* slave_loop*/
#endif


