/*
 * mpislavequeue.c
 *
 *
 * Part of TREE-PUZZLE 5.2 (February 2005)
 *
 * (c) 2003-2005 by Heiko A. Schmidt, Korbinian Strimmer, and Arndt von Haeseler
 * (c) 1999-2003 by Heiko A. Schmidt, Korbinian Strimmer,
 *                  M. Vingron, and Arndt von Haeseler
 * (c) 1995-1999 by Korbinian Strimmer and Arndt von Haeseler
 *
 * All parts of the source except where indicated are distributed under
 * the GNU public licence.  See http://www.opensource.org for details.
 *
 * ($Id: sched.c,v 1.1 2005/09/29 13:32:18 hschmidt Exp $)
 *
 */



#if 0
# define EXTERN extern
#endif
 
int *freeslaves;           /* Queue of free slaves */
int firstslave,            /* headpointer of queue */
    lastslave;             /* tailpointer of queue */

int PI_MaxSlave = 0;

void PI_Finalize(void);

/*********************************************************************
*********************************************************************/

void sq_maerror(char *message)
{
        fprintf(stdout, "\n\n\nUnable to proceed (lack of memory: %s)\n\n", message);
#       if USE_MPI
                PI_Finalize();
#       endif
        exit(1);
} /* sq_maerror */

/******************/

int *sq_new_ivector(int n)
{
        int *v;

        v = (int *) calloc((size_t) n, sizeof(int));
        if (v == NULL) sq_maerror("step 1 in new_ivector");

        return v;
} /* new_ivector */

/******************/

void sq_free_ivector(int *v)
{
        free((int *) v);
} /* sq_free_ivector */


/******************/


/*********************************************************************
*  queue for storing the ranks of slaves waiting for work            *
*********************************************************************/

void PI_initslavequeue(int NumProcs)
{
  int n;
  freeslaves = sq_new_ivector(NumProcs);
    firstslave = 0;
    PI_MaxSlave = NumProcs-1;
    lastslave  = PI_MaxSlave-1;
    freeslaves[PI_MaxSlave] = PI_MaxSlave;
    for (n=0; n<PI_MaxSlave; n++){
      freeslaves[n] = n+1;
    }
} /* PI_initslavequeue */

/******************/

int PI_fullslave(void)
{
return (freeslaves[PI_MaxSlave] == PI_MaxSlave);
} /* PI_fullslave */

/******************/

int PI_emptyslave(void)
{
return (freeslaves[PI_MaxSlave] == 0);
} /* PI_emptyslave */

/******************/

void PI_putslave(int sl)
{
  if (freeslaves[PI_MaxSlave] == PI_MaxSlave) {
    fprintf(STDOUT, "\n\n\nHALT: PLEASE REPORT ERROR PP1 TO DEVELOPERS\n\n\n");
    MPI_Finalize();
    exit(1);
  }
  (freeslaves[PI_MaxSlave])++;
  lastslave = (lastslave + 1) % PI_MaxSlave;
  freeslaves[lastslave] = sl;
} /* PI_putslave */

/******************/

int PI_getslave(void)
{
  int tmp;
  if (freeslaves[PI_MaxSlave] == 0)
    return MPI_PROC_NULL;
  else {
    tmp = freeslaves[firstslave];
    firstslave = (firstslave + 1) % PI_MaxSlave;
    (freeslaves[PI_MaxSlave])--;
    return tmp;
  }
} /* PI_getslave */

