/*
 * sub1.c
 *
 *
 * Part of (p)IPHULA 1.16 (May 2007)
 *
 * (c) 2004-2007 by Heiko A. Schmidt, Jutta Buschbom, 
 *                  Gunter Weiss, and Arndt von Haeseler
 * (c) 1998-2004 by Gunter Weiss and Arndt von Haeseler
 *
 * All parts of the source except where indicated are distributed 
 * under the GNU public licence 2. 
 * See http://www.opensource.org/licenses/gpl-license.php details.
 *
 * ($Id: sub1.c,v 1.11 2005/09/30 01:23:17 hschmidt Exp $)
 *
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>

#include <stddef.h>

#ifdef _OPENMP
#	define USE_OPENMP 1
#	include "omp.h"
#else
#       ifdef USE_OPENMP
#               undef USE_OPENMP
#       endif
#endif

#ifdef MPI
#       define USE_MPI 1
#       include <mpi.h>
#else
#       ifdef USE_MPI
#               undef USE_MPI
#       endif
#endif


#define EXTERN extern
#include "iphula.h"
#include "sub1.h"
#include "sub2.h"
#include "config.h"

#ifdef NOSPRNG
#   ifdef SPRNG
#      undef SPRNG

#   endif
#else
#   define SPRNG 1
#endif

#ifdef SPRNG
#  include "sprng/sprng.h"
#endif

#define NR_END 1
#define FREE_ARG char*

#if 0
#define RANDVERBOSE1 1
#endif
 
#define PI 3.141592654

#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

/* variables  */

node *Ctree;

int NumSamples;		/* number of taxa ??? (HAS) */
long NumRuns;		/* number of repititions */
TTree *tree;
TNode **Tlist;

/* THETA: the genetic diversity at equilibrium          */
/*        in the past (=2*pop.size*mut.rate)            */
/* TAU: the time back when population started           */
/*      to change in size (measured in mut.time units)  */
/* RHO: ratio between pop.sizes of today and before tau */
double MinTheta = 1.0;
double MaxTheta = 10.0;
double MinTau = 0.5;
double MaxTau = 5;
/* change ;-) - done */
/* double Rho; */ /* obsoleted by Rho_arr */
double Rho_arr[MAX_RHO_STEPS]; /* def'ed external in sub1.h */
int Rhosteps=8;
int Thetasteps = 10;
int Tausteps = 10;

/* Opening Monolog */
void OpenDialog(void)
{
#if 0
	printf("\n\n ******       IPHULA %s      ******\n\n", VERSION);
#endif
#	ifdef USE_MPI
#		ifdef USE_OPENMP
			fprintf(stdout,"\n\n******   IPHULA %s (hybrid MPI/OpenMP binary)   ******\n\n", VERSION); fflush(stdout);
#		else
			fprintf(stdout,"\n\n******   IPHULA %s (MPI-parallel binary)   ******\n\n", VERSION); fflush(stdout);
#		endif
#	else
#		ifdef USE_OPENMP
			fprintf(stdout,"\n\n******   IPHULA %s (OpenMP-parallel binary)   ******\n\n", VERSION); fflush(stdout);
#		else
			fprintf(stdout,"\n\n******   IPHULA %s (sequential binary)   ******\n\n", VERSION); fflush(stdout);
#		endif
#	endif
	fprintf(stdout, "This program makes inference of population history\n");
	fprintf(stdout, "parameters based on the likelihood of a parameter set\n");
	fprintf(stdout, "given the number of variable positions in the sample\n");
	fprintf(stdout, "and the mean pairwise distance.\n");
	fprintf(stdout, "For more details and reference issues see:\n");
	fprintf(stdout, "G.Weiss & A.v.Haeseler (1998) Genetics:149, pp.1539-1546\n"); fflush(stdout);
}


/* setting: Freq[0..3], kappa(ts/tv ratio), ksi(pyr/pur ts ratio), GammaShape */
void SubsProDialog(FILE* infp, FILE* outfp)
{
	char c;
	int dummy;
	double freqsum;

	printf("\n");
	printf("Current parameters of the substitution process:\n");
	printf("-----------------------------------------------\n\n");
	printf("base frequency of A  : %.3f\n",Freq[0]);
	printf("                  C  : %.3f\n",Freq[1]);
	printf("                  G  : %.3f\n",Freq[2]);
	printf("                  T  : %.3f\n",Freq[3]);

	printf("ts/tv ratio kappa    : %.1f\n",kappa);
	printf("pyr/pur ts ratio xi  : %.2f\n",ksi);

	if (NumCats > 1) {
		printf("rate heterogeity     : yes\n");
		printf("Gamma rate categories: %d\n",NumCats);
		printf("shape parameter alpha: %.2f\n",GammaShape);
	} else {
		printf("rate heterogeity     : no\n");
	}

	printf("\nDo you want to accept these values for the substitution process?\n");
	printf("type 'y' to accept, 'n' to change, or 'q' to abort: "); fflush(stdout);
	fscanf(infp, "%c", &c);
	while (getc(infp) != '\n'); /* Remove until end of line */

	if((c == 'q') || (c == 'Q')) {
		exit(0);
	}

	if((c != 'y') && (c != 'Y')) {

		printf("\nbase frequency of A: ");
		dummy = fscanf(infp, "%lf", &Freq[0]);
		if (dummy != 1) {
			fprintf(stderr, "INPUT ERROR: base frequency of A!\n");
			fprintf(outfp, "INPUT ERROR: base frequency of A!\n");
			fclose(outfp);
			exit(1);
		}
		while (getc(infp) != '\n'); /* Remove until end of line */

		printf("\nbase frequency of C: ");
		dummy = fscanf(infp, "%lf", &Freq[1]);
		if (dummy != 1) {
			fprintf(stderr, "INPUT ERROR: base frequency of C!\n");
			fprintf(outfp, "INPUT ERROR: base frequency of C!\n");
			fclose(outfp);
			exit(1);
		}
		while (getc(infp) != '\n'); /* Remove until end of line */

		printf("\nbase frequency of G: ");
		dummy = fscanf(infp, "%lf", &Freq[2]);
		if (dummy != 1) {
			fprintf(stderr, "INPUT ERROR: base frequency of G!\n");
			fprintf(outfp, "INPUT ERROR: base frequency of G!\n");
			fclose(outfp);
			exit(1);
		}
		while (getc(infp) != '\n'); /* Remove until end of line */

		printf("\nbase frequency of T: ");
		dummy = fscanf(infp, "%lf", &Freq[3]);
		if (dummy != 1) {
			fprintf(stderr, "INPUT ERROR: base frequency of T!\n");
			fprintf(outfp, "INPUT ERROR: base frequency of T!\n");
			fclose(outfp);
			exit(1);
		}
		while (getc(infp) != '\n'); /* Remove until end of line */

		freqsum = Freq[0]+Freq[1]+Freq[2]+Freq[3];
		if(freqsum != 1.0) {
		  Freq[0] /= freqsum;
		  Freq[1] /= freqsum;
		  Freq[2] /= freqsum;
		  Freq[3] /= freqsum;
		  printf("WARNING: Frequencies normalized, because sum not 1.0!\n");
		  fprintf(outfp, "WARNING: Frequencies normalized, because sum not 1.0!\n");
		}

		printf("\nts/tv ratio   kappa: ");
		dummy = fscanf(infp, "%lf", &kappa);
		if (dummy != 1) {
			fprintf(stderr, "INPUT ERROR: nts/tv ratio kappa!\n");
			fprintf(outfp, "INPUT ERROR: nts/tv ratio kappa!\n");
			fclose(outfp);
			exit(1);
		}
		while (getc(infp) != '\n'); /* Remove until end of line */

		printf("\npyr/pur ts ratio xi: ");
		dummy = fscanf(infp, "%lf", &ksi);
		if (dummy != 1) {
			fprintf(stderr, "INPUT ERROR: npyr/pur ts ratio xi!\n");
			fprintf(outfp, "INPUT ERROR: npyr/pur ts ratio xi!\n");
			fclose(outfp);
			exit(1);
		}
		while (getc(infp) != '\n'); /* Remove until end of line */

		printf("\nnumber of rate categories [1=homogeneous]: ");
		dummy = fscanf(infp, "%d", &NumCats);
		if (dummy != 1) {
			fprintf(stderr, "INPUT ERROR: number of rate categories!\n");
			fprintf(outfp, "INPUT ERROR: number of rate categories!\n");
			fclose(outfp);
			exit(1);
		}
		while (getc(infp) != '\n'); /* Remove until end of line */

		if (NumCats > 1)  {
			printf("\nGamma shape parameter alpha: ");
			dummy = fscanf(infp, "%lf", &GammaShape);
			if (dummy != 1) {
				fprintf(stderr, "INPUT ERROR: Gamma shape parameter alpha!\n");
				fprintf(outfp, "INPUT ERROR: Gamma shape parameter alpha!\n");
			fclose(outfp);
			exit(1);
			}
			while (getc(infp) != '\n'); /* Remove until end of line */
		}

		SubsProDialog(infp, outfp);
	}

} /* SubsProDialog */



/*********************/

/* setting: NumSamples, NumBases, NumVarPos, MeanPWDist */
void DataDialog(FILE *infp, FILE* outfp)
{
	char cc;
	int dummy;
	/* int ok; */ /* HAS ;-) */

	printf("\n\n");
	printf("Please specify your data set: \n");
	printf("----------------------------- \n");

	printf("\nsample size (number of sequences): ");		/* number of taxa ??? (HAS) */
	dummy = fscanf(infp, "%d", &NumSamples);		/* number of taxa ??? (HAS) */
	if (dummy != 1) {
		fprintf(stderr, "INPUT ERROR: sample size (number of sequences)!\n");
		fprintf(outfp, "INPUT ERROR: sample size (number of sequences)!\n");
		fclose(outfp);
		exit(1);
	}
	while (getc(infp) != '\n'); /* Remove until end of line */

	if (NumSamples < 2) {
		fprintf(stderr, "ERROR: The sample size needs to be >= 2\n");
		fprintf(outfp, "ERROR: The sample size needs to be >= 2\n");
		fclose(outfp);
		exit(1);
		NumSamples = 10;
	}

	printf("\nsequence length (bases): ");
	dummy = fscanf(infp, "%d", &NumBases);
		if (dummy != 1) {
			fprintf(stderr, "INPUT ERROR: sequence length (bases)!\n");
			fprintf(outfp, "INPUT ERROR: sequence length (bases)!\n");
		fclose(outfp);
		exit(1);
		}
		while (getc(infp) != '\n'); /* Remove until end of line */

	if (NumBases <= 1) {
		fprintf(stderr, "ERROR: The sequence length needs to be > 1\n");
		fprintf(outfp, "ERROR: The sequence length needs to be > 1\n");
		fclose(outfp);
		exit(1);
		NumBases = 1000;
	}

	printf("\nnumber of variable positions: ");
	dummy = fscanf(infp, "%d", &NumVarPos);
	if (dummy != 1) {
		fprintf(stderr, "INPUT ERROR: number of variable positions!\n");
		fprintf(outfp, "INPUT ERROR: number of variable positions!\n");
		fclose(outfp);
		exit(1);
	}
	while (getc(infp) != '\n'); /* Remove until end of line */

	if (NumVarPos < 0) {
		NumVarPos = (NumBases>=10 ? 10 : NumBases);
		fprintf(stderr, "ERROR: The number of variable sites needs to be >= 0\n");
		fprintf(outfp, "ERROR: The number of variable sites needs to be >= 0\n");
		fclose(outfp);
		exit(1);
	}
	if (NumVarPos > NumBases) {
		fprintf(stderr, "ERROR: Number of variable sites larger than sequence length\n");
		fprintf(outfp, "ERROR: Number of variable sites larger than sequence length\n");
		fclose(outfp);
		exit(1);
		NumVarPos = NumBases;
	}

	printf("\nmean pairwise distance: ");
	dummy = fscanf(infp, "%lf", &MeanPWDist);
	if (dummy != 1) {
		fprintf(stderr, "INPUT ERROR: mean pairwise distance!\n");
		fprintf(outfp, "INPUT ERROR: mean pairwise distance!\n");
		fclose(outfp);
		exit(1);
	}
	while (getc(infp) != '\n'); /* Remove until end of line */

	if(MeanPWDist < 0.0) {
		MeanPWDist = 0.1;
		fprintf(stderr, "ERROR: The mean pairwise distance needs to be >= 0\n");
		fprintf(outfp, "ERROR: The mean pairwise distance needs to be >= 0\n");
		fclose(outfp);
		exit(1);
	}

	printf("\n\nInformation from your data set:\n");
	printf("------------------------------:\n\n");
	printf("sample size (# sequences) : %d\n",NumSamples);
	printf("sequence length (bases)   : %d\n",NumBases);
	printf("variable positions        : %d\n",NumVarPos);
	printf("mean pairwise distance    : %.2f\n",MeanPWDist);

	dummy = printf("\nIs this correct? Type 'y' to accept, 'n' to change, or 'q' to abort: "); fflush(stdout);
	fscanf(infp, "%c", &cc);
	while (getc(infp) != '\n'); /* Remove until end of line */

	if((cc == 'q') || (cc == 'Q')) {
		exit(0);
	}

	if((cc != 'y') && (cc != 'Y')) {
		DataDialog(infp, outfp);
	}

} /* DataDialog */

/*********************************/

/* setting: Rhosteps, Rho_arr, (Rho,) MinTheta, MaxTheta, Thetasteps, MinTau, MaxTau, Tausteps, NumRuns */
void SimParDialog(FILE *infp, FILE* outfp)
{
	char   c, cc;
	int    dummy;
	int    n;			/* counter variable */
	double tmpRho_arr[MAX_RHO_STEPS];
	int    tmpRhosteps;
	double tmpMinTau, tmpMaxTau;
	int    tmpTausteps;
	double tmpMinTheta, tmpMaxTheta;
	int    tmpThetasteps;
	int    no_tau = FALSE;

	tmpRhosteps = Rhosteps;
	for (n=0; n<Rhosteps; n++)
		tmpRho_arr[n] = Rho_arr[n];

	tmpTausteps = Tausteps;
	tmpMaxTau = MaxTau;
	tmpMinTau = MinTau;

	tmpThetasteps = Thetasteps;
	tmpMaxTheta = MaxTheta;
	tmpMinTheta = MinTheta;

	printf("\n\nModel of population history:\n");
	printf("----------------------------\n");
	printf("\nThe model of population history used contains three parameters:\n");
	printf("- THETA: the genetic diversity at equilibrium \n");
	printf("         in the past (=2*pop.size*mut.rate) \n");
	printf("- TAU:   the time back when population started \n");
	printf("         to change in size (measured in mut.time units)\n");
	printf("- RHO:   the ratio between population sizes of today and before TAU.\n");
	printf("Note: TAU remains undefined if RHO = 1");

#if 0
	/* printf("\n\nFirst specify RHO: "); */
	printf("\n\nFirst specify RHO: ");
/* change ;-) - if 0 */
	dummy = fscanf(infp, "%lf",&Rho);
	if (dummy != 1) {
		fprintf(stderr, "INPUT ERROR: RHO!\n");
		fprintf(outfp, "INPUT ERROR: RHO!\n");
		fclose(outfp);
		exit(1);
	}
	while (getc(infp) != '\n'); /* Remove until end of line */

	printf("\nnumber of RHO grid points: ");
	dummy = fscanf(infp, "%d",&Rhosteps);
	if (dummy != 1) {
		fprintf(stderr, "INPUT ERROR: number of RHO grid points!\n");
		fprintf(outfp, "INPUT ERROR: number of RHO grid points!\n");
		fclose(outfp);
		exit(1);
	}
	while (getc(infp) != '\n'); /* Remove until end of line */

/* change ;-) - if 0 */
	if((0.99<Rho) && (Rho<1.01)) { /* is RHO == 1.0 */
		no_tau = TRUE;
		Tausteps = 1;
		MaxTau = MinTau = 0.0;

	} else {
		no_tau = FALSE;
	}
#endif

	printf("\n\nPopulation history parameter:\n");
	printf("-----------------------------\n");
	if (no_tau && Rhosteps==1) {
		printf("\nThe simulations will run on a grid of RHO and THETA (TAU undefined) having\n");
		printf("- RHO   with 1 grid point: %g\n", Rho_arr[0]);
		printf("- THETA with %3d grid points from %5.1f to %5.1f. \n", 
			Thetasteps, MinTheta, MaxTheta);
	} else {
		printf("\nThe simulations will run on a grid of RHO, THETA and TAU having\n");
		printf("- RHO   with %3d grid points: \n\t", Rhosteps);
		for (n=0; n<Rhosteps; n++)
			printf("%s %g", (n==0 ? "" : ","), Rho_arr[n]);
		printf("\n");
		printf("- THETA with %3d grid points from %5.1f to %5.1f \n", 
			Thetasteps, MinTheta, MaxTheta);
		printf("- TAU   with %3d grid points from %5.1f to %5.1f \n", 
			Tausteps, MinTau, MaxTau);
	}
	printf("\nSimulate with these grid parameters?");
	printf("\nType 'y' to accept, 'n' to change, or 'q' to abort: "); fflush(stdout);
	dummy = fscanf(infp, "\n%c", &c);
	/* if (dummy != 1) printf("INPUT ERROR - old value kept!\n"); */
	while (getc(infp) != '\n'); /* Remove until end of line */

	if((c == 'q') || (c == 'Q')) {
		exit(0);
	}

	if((c != 'y') && (c != 'Y')) {
		/*** Reading Rho ***/
		printf("\nnumber of RHO grid points: ");
		dummy = fscanf(infp, "%d",&Rhosteps);
		if (dummy != 1) {
			fprintf(stderr, "INPUT ERROR: number of RHO grid points!\n");
			fprintf(outfp, "INPUT ERROR: number of RHO grid points!\n");
			fclose(outfp);
			exit(1);
		}
		while (getc(infp) != '\n');  /* Remove until end of line */

		if (Rhosteps < 0) {
			fprintf(stderr, "INPUT ERROR: Number of Rho grid points must be > 0!\n\n");
			fprintf(outfp, "INPUT ERROR: Number of Rho grid points must be > 0!\n\n");
			fclose(outfp);
			exit(1);
			Rhosteps = tmpRhosteps;
			for (n=0; n<tmpRhosteps; n++)
				Rho_arr[n] = tmpRho_arr[n];
		}

		if (Rhosteps > MAX_RHO_STEPS) {
			fprintf(stderr, "INPUT ERROR: Number of Rho grid points must be < %d!\n(You can extend this bound by increasing MAX_RHO_STEPS in sub1.h.)\n\n", MAX_RHO_STEPS + 1);
			fprintf(outfp, "INPUT ERROR: Number of Rho grid points must be < %d!\n(You can extend this bound by increasing MAX_RHO_STEPS in sub1.h.)\n\n", MAX_RHO_STEPS + 1);
			fclose(outfp);
			exit(1);
			Rhosteps = tmpRhosteps;
			for (n=0; n<tmpRhosteps; n++)
				Rho_arr[n] = tmpRho_arr[n];
		}

		for (n=0; n<Rhosteps; n++) {
			printf("\n%2d. RHO: ", n+1);
			dummy = fscanf(infp, "%lf", &Rho_arr[n]);
			if (dummy != 1) {
				fprintf(stderr, "INPUT ERROR: min RHO!\n");
				fprintf(outfp, "INPUT ERROR: min RHO!\n");
				fclose(outfp);
				exit(1);
			}
			if (Rho_arr[n] <= 0.0) {
				fprintf(stderr, "INPUT ERROR: Rho must be > 0!\n\n", MAX_RHO_STEPS + 1);
				fprintf(outfp, "INPUT ERROR: Rho must be > 0!\n\n", MAX_RHO_STEPS + 1);
				fclose(outfp);
				exit(1);
				Rhosteps = tmpRhosteps;
				for (n=0; n<tmpRhosteps; n++)
					Rho_arr[n] = tmpRho_arr[n];
			}
			if((0.99<Rho_arr[n]) && (Rho_arr[n]<1.01)) { /* is RHO == 1.0 */
				Rho_arr[n] = 1.0;
			}

			while (getc(infp) != '\n'); /* Remove until end of line */
		}
		if((Rhosteps == 1) && (Rho_arr[0] == 1.0)) { /* single RHO == 1.0 */
			no_tau = TRUE;
			Tausteps = 1;
			MaxTau = MinTau = 0.0;
	
		} else {
			no_tau = FALSE;
		}
		/*** END Reading Rho ***/

		/*** Reading Theta ***/
		printf("\nnumber of THETA grid points: ");
		dummy = fscanf(infp, "%d",&Thetasteps);
		if (dummy != 1) {
			fprintf(stderr, "INPUT ERROR: number of THETA grid points!\n");
			fprintf(outfp, "INPUT ERROR: number of THETA grid points!\n");
			fclose(outfp);
			exit(1);
		}
		while (getc(infp) != '\n'); /* Remove until end of line */

		printf("\nmin THETA: ");
		dummy = fscanf(infp, "%lf", &MinTheta);
		if (dummy != 1) {
			fprintf(stderr, "INPUT ERROR: min THETA!\n");
			fprintf(outfp, "INPUT ERROR: min THETA!\n");
			fclose(outfp);
			exit(1);
		}
		while (getc(infp) != '\n'); /* Remove until end of line */

		printf("\nmax THETA: ");
		dummy = fscanf(infp, "%lf", &MaxTheta);
		if (dummy != 1) {
			fprintf(stderr, "INPUT ERROR: max THETA!\n");
			fprintf(outfp, "INPUT ERROR: max THETA!\n");
			fclose(outfp);
			exit(1);
		}
		while (getc(infp) != '\n'); /* Remove until end of line */

		if ((0 > MinTheta) || (MinTheta > MaxTheta) || (Thetasteps < 1)) {
			fprintf(stderr, "INPUT ERROR: Entries need to be 0 < min <= max!\n\n");
			fprintf(outfp, "INPUT ERROR: Entries need to be 0 < min <= max!\n\n");
			fclose(outfp);
			exit(1);
			Thetasteps = tmpThetasteps;
			MaxTheta = tmpMaxTheta;
			MinTheta = tmpMinTheta;
		}
		/*** END Reading Theta ***/


		/*** Reading Tau ***/
		if (no_tau != TRUE) {
			printf("\nnumber of TAU grid points: ");
			dummy = fscanf(infp, "%d",&Tausteps);
			if (dummy != 1) {
				fprintf(stderr, "INPUT ERROR: number of TAU grid points!\n");
				fprintf(outfp, "INPUT ERROR: number of TAU grid points!\n");
				fclose(outfp);
				exit(1);
			}
			while (getc(infp) != '\n'); /* Remove until end of line */

			printf("\nmin TAU: ");
			dummy = fscanf(infp, "%lf", &MinTau);
			if (dummy != 1) {
				fprintf(stderr, "INPUT ERROR: min TAU!\n");
				fprintf(outfp, "INPUT ERROR: min TAU!\n");
				fclose(outfp);
				exit(1);
			}
			while (getc(infp) != '\n'); /* Remove until end of line */

			printf("\nmax TAU: ");
			dummy = fscanf(infp, "%lf", &MaxTau);
			if (dummy != 1) {
				fprintf(stderr, "INPUT ERROR: max TAU!\n");
				fprintf(outfp, "INPUT ERROR: max TAU!\n");
				fclose(outfp);
				exit(1);
			}
			while (getc(infp) != '\n'); /* Remove until end of line */

			if (0 > MinTau || MinTau > MaxTau || Tausteps < 1) {
				fprintf(stderr, "INPUT ERROR: Entries need to be 0 < min <= max!\n\n");
				fprintf(outfp, "INPUT ERROR: Entries need to be 0 < min <= max!\n\n");
				fclose(outfp);
				exit(1);
				Tausteps = tmpTausteps;
				MaxTau = tmpMaxTau;
				MinTau = tmpMinTau;
			}
		}
		/*** END Reading Tau ***/

#if 0
		printf("\n\nEntered gid parameters for THETA and TAU:\n");
		printf("Values for genetic diversity THETA\n");
		printf("   from min:  %.1f\n", MinTheta);
		printf("   to max:    %.1f\n", MaxTheta);
		printf("   with %d grid points\n\n", Thetasteps);
		printf("Values for time TAU\n");
		printf("   from min:  %.1f\n", MinTau);
		printf("   to max:    %.1f\n", MaxTau);
		printf("   with %d grid points", Tausteps);

		printf("\nIf this is correct, type 'y', otherwise 'n': ");
		dummy = fscanf(infp, "\n%c", &cc);
		if (dummy != 1) printf("INPUT ERROR - old value kept!\n");
		while (getc(infp) != '\n'); /* Remove until end of line */

		if(cc != 'y') SimParDialog();
#endif 
	}

	/* } */

	printf("\n\nFinally, specify the number of simulations on each grid point: ");
	dummy = fscanf(infp, "%ld", &NumRuns);
	if (dummy != 1) {
		fprintf(stderr, "INPUT ERROR: number of simulations on each grid point!\n");
		fprintf(outfp, "INPUT ERROR: number of simulations on each grid point!\n");
		fclose(outfp);
		exit(1);
	}
	while (getc(infp) != '\n'); /* Remove until end of line */

	if (NumRuns < 1) {
		NumRuns = 1000;
		fprintf(stderr, "WARNING: The number of simulations must be numeric and equal to or larger than 1! (set to %ld)\n", NumRuns);
		fprintf(outfp, "WARNING: The number of simulations must be numeric and equal to or larger than 1! (set to %ld)\n", NumRuns);
	}

	printf("\nand the tolerance limit delta: ");
	dummy = fscanf(infp, "%lf", &tolerance);
	if (dummy != 1) {
		fprintf(stderr, "INPUT ERROR: tolerance limit delta!\n");
		fprintf(outfp, "INPUT ERROR: tolerance limit delta!\n");
		fclose(outfp);
		exit(1);
	}
	while (getc(infp) != '\n'); /* Remove until end of line */

	if (tolerance <= 0.0) {
		tolerance = 0.2;
		fprintf(stderr, "The tolerance needs to be larger than zero!\n");
		fprintf(outfp, "The tolerance needs to be larger than zero!\n");
		fclose(outfp);
		exit(1);
	}

	printf("\n%ld simulations per grid point \nwith a tolerance of %.2f will be run\n", NumRuns, tolerance);


	printf("\n\nCurrent simulation parameters for RHO, THETA, and TAU:\n");
	printf("- Population size ratios RHO: %g", Rho_arr[0]);
	for (n=1; n<Rhosteps; n++) {
		printf(",\n                              %g", Rho_arr[n]);
		if (Rho_arr[n] == 1.0) printf(" (TAU undefined)");
	}
	printf("\n       number of grid points: %d\n\n", Rhosteps);
	printf("- Values for THETA  from min: %.6f\n", MinTheta);
	printf("                      to max: %.6f\n", MaxTheta);
	printf("       number of grid points: %d\n", Thetasteps);
	if (no_tau == TRUE) {
		printf("- TAU is undefined (RHO = 1)\n\n");
	} else {
		printf("- Values for TAU    from min: %.6f\n", MinTau);
		printf("                      to max: %.6f\n", MaxTau);
		printf("       number of grid points: %d\n", Tausteps);
	}
	printf("\n- simulations per grid point: %ld\n", NumRuns);
	printf("- tolerance limit      delta: %.4f\n", tolerance);

	printf("\nIs this correct? Type 'y' to accept, 'n' to change, or 'q' to abort: "); fflush(stdout);
	dummy = fscanf(infp, "\n%c", &cc);
	while (getc(infp) != '\n'); /* Remove until end of line */

	if((cc == 'q') || (cc == 'Q')) {
		exit(0);
	}

	if((cc != 'y') && (cc != 'Y')) {
		SimParDialog(infp, outfp);
	}

#	ifdef VERBOSE
		fprintf(stderr, "\nthetasteps=%d, mintheta=%f, maxtheta=%f\n", Thetasteps, MinTheta, MaxTheta);
#	endif /* VERBOSE */

	if ((Thetasteps==1) &&  (MinTheta != MaxTheta)) {
		fprintf(stderr, "ERROR: min Theta and max Theta must be equal, if number of Theta grid points = 1\n");
		fprintf(outfp, "ERROR: min Theta and max Theta must be equal, if number of Theta grid points = 1\n");
		fclose(outfp);
		exit(1);
	}

	if ((Tausteps==1) &&  (MinTau != MaxTau)) {
		fprintf(stderr, "ERROR: min Tau and max Tau must be equal, if number of Tau grid points = 1\n");
		fprintf(outfp, "ERROR: min Tau and max Tau must be equal, if number of Tau grid points = 1\n");
		fclose(outfp);
		exit(1);
	}

} /* SimParDialog */
/*********************************/

/********************************/

#ifdef NO_SPRNG
	static long idum;

	long seedran(long randseed)
	{
		if (randseed>0)
			idum=-randseed;
		else 
			idum=randseed;
		return idum;
	}


	/* draw random number 0 <= x < 1 */
	/* double ran2(void) */
	double randdouple(int dummy)
	{
		int j;
		long k;
		static long idum2=123456789;
		static long iy=0;
		static long iv[NTAB];
		double temp;

		if (idum <= 0) {
			if (-(idum) < 1) idum=1;
			else idum = -(idum);
			idum2=(idum);
			for (j=NTAB+7;j>=0;j--) {
				k=(idum)/IQ1;
				idum=IA1*(idum-k*IQ1)-k*IR1;
				if (idum < 0) idum += IM1;
				if (j < NTAB) iv[j] = idum;
			}
			iy=iv[0];
		}
		k=(idum)/IQ1;
		idum=IA1*(idum-k*IQ1)-k*IR1;
		if (idum < 0) idum += IM1;
		k=idum2/IQ2;
		idum2=IA2*(idum2-k*IQ2)-k*IR2;
		if (idum2 < 0) idum2 += IM2;
		j=iy/NDIV;
		iy=iv[j]-idum2;
		iv[j] = idum;
		if (iy < 1) iy += IMM1;
		if ((temp=AM*iy) > RNMX) return RNMX;
		else return temp;
	}
	#undef IM1
	#undef IM2
	#undef AM
	#undef IMM1
	#undef IA1
	#undef IA2
	#undef IQ1
	#undef IQ2
	#undef IR1
	#undef IR2
	#undef NTAB
	#undef NDIV
	#undef EPS
	#undef RNMX

#else /* SPRNG */

	int **randstream;

	/* int initrandom(long randseed) */
	long seedran(long randseed)
	{
		int n;
		srand((unsigned) time(NULL));
		if (randseed < 0) 
			randseed = (long)make_sprng_seed();

		randstream = (int **)malloc(sizeof(int *) * NumThreads);
		for(n=0; n<NumThreads; n++) randstream[n] = NULL;

#		ifndef USE_OPENMP
			fprintf(stdout,"Using SPRNG -- Scalable (Parallel) Random Number Generator\n");
			fflush(stdout);
			randstream[0] = init_sprng(0,1,(int)randseed,SPRNG_DEFAULT); /*init stream*/

#			ifdef RANDVERBOSE1
				fprintf(stdout, "!!! random seed set to %d !!!\n", randseed);
				printf("Printing information about new stream\n");
				print_sprng(randstream[0]);
#			endif /* RANDVERBOSE 1 */
#		else /* ! USE_OPENMP */
			fprintf(stdout,"Using SPRNG -- Scalable Parallel Random Number Generator\n");
			fflush(stdout);
			/* MPI_Bcast(&randseed, 1, MPI_UNSIGNED, PP_MyMaster, MPI_COMM_WORLD); */
			for(n=0; n<NumThreads; n++) {
				randstream[n] = init_sprng(n,NumThreads,(int)randseed,SPRNG_DEFAULT); /*initialize stream*/
#				ifdef RANDVERBOSE1
					fprintf(stdout, "(%2d) !!! random seed set to %d !!!\n", MyThreadNum, randseed);
					printf(" Printing information about new stream\n");
					print_sprng(randstream[n]);
#				endif  /* RANDVERBOSE1 */
			}
#		endif /* ! USE_OPENMP */
		return (randseed);
	}  /* initrandom */ 



	/* draw random number 0 <= x < 1 */
	double randdouple(int threadnum)
	{
		return sprng(randstream[threadnum]);
	}
#endif



/* TO BE MADE THREAD-SAFE (HAS) */
/* new external  variables: NumSamples,  */
/* new internal  variables: i, outTlist, outtree, */
/* new parameter variables: Ctree, otree->&tree, oTlist->&Tlist */

/* old external  variables: sample, Tlist,  */
/* old internal  variables: i */
/* old parameter variables: Ctree, tree */
/* calling:    */
/* called by: grow_tree */
void C_2_Ttree(node *Ctree, TTree **otree, TNode ***oTlist)
{
	int i;
	TNode **outTlist;
	TTree *outtree;

	outTlist = (TNode **)malloc((2*NumSamples-1)*sizeof(TNode *));
	for(i=0;i<2*NumSamples-1;i++)
		outTlist[i] = (TNode *)malloc(sizeof(TNode));	

	outtree=*otree;
	outtree->rooted = 1;
	outtree->lengths = -1;
	outtree->root = outTlist[2*NumSamples-2];
	outtree->nodeList = outTlist[2*NumSamples-2];
	outtree->numTips = NumSamples;
	outtree->numNodes = 2*NumSamples-1;
	outtree->totalLength = Ctree[2*NumSamples - 2].time;
	for(i=0;i<NumSamples; i++)
		outtree->tips[i] = outTlist[i];


	for(i=2*NumSamples-2;i>=NumSamples; i--)
	{
		outTlist[i]->branch1 = outTlist[Ctree[i].desc1index];
		outTlist[i]->branch2 = outTlist[Ctree[i].desc2index];
		outTlist[i]->branch0 = NULL;
		outTlist[i]->next = outTlist[i-1];
		outTlist[i]->tipNo = -1;
	}


	for(i=NumSamples-1;i>0; i--)
	{
		outTlist[i]->branch1 = NULL;
		outTlist[i]->branch2 = NULL;
		outTlist[i]->branch0 = NULL;
		outTlist[i]->next = outTlist[i-1];
		outTlist[i]->length0 = (Ctree[i].ancestor->time) - (Ctree[i].time);
		outTlist[i]->length1 = 0;
		outTlist[i]->length2 = 0;
		outTlist[i]->param = 0;
		outTlist[i]->tipNo = i;
	}

	outTlist[0]->branch1 = NULL;
	outTlist[0]->branch2 = NULL;
	outTlist[0]->branch0 = NULL;
	outTlist[0]->next = NULL;
	outTlist[0]->length0 = (Ctree[0].ancestor->time) - (Ctree[0].time);
	outTlist[0]->length1 = 0;
	outTlist[0]->length2 = 0;
	outTlist[0]->param = 0;
	outTlist[0]->tipNo = 0;


	outTlist[2*NumSamples-2]->length0 = 0;
	outTlist[2*NumSamples-2]->length1 = outTlist[2*NumSamples-2]->branch1->length0;
	outTlist[2*NumSamples-2]->length2 = outTlist[2*NumSamples-2]->branch2->length0;
	outTlist[2*NumSamples-2]->param = 0;

	for(i=2*NumSamples-3;i>=NumSamples; i--)
	{
		outTlist[i]->length0 = (Ctree[i].ancestor->time) - (Ctree[i].time);
		outTlist[i]->length1 = outTlist[i]->branch1->length0;
		outTlist[i]->length2 = outTlist[i]->branch2->length0;
		outTlist[i]->param = 0;
	}

	*oTlist = outTlist;

} /* C_2_Ttree */



/* TO BE MADE THREAD-SAFE (HAS) */
/* new external  variables: NumSamples(=NumTaxa), */
/* new internal  variables: in, pick, list, Ctree, Croot, t, t_i, s, rr, L_t, L_s */
/* new parameter variables: Tlist, tree, factor, Tt */

/* old external  variables: tree, Ctree, sample,  */
/* old internal  variables: in, pick, list, Croot, t, t_i, s, rr, L_t, L_s */
/* old parameter variables: factor, Tt */
/* calling:   C_2_Ttree, log, randdouple, exp */
/* called by: main */
void grow_tree(double factor, double Tt, TTree **outtreeptr, TNode ***outTlistptr)
{
	int in, pick;
	node **list;
	/* node *Croot; (HAS) */
	double t, t_i, s, rr;
	double L_s;
	/* double L_t; (HAS) */

	node *Ctree;

	/* Initialize things */

	Ctree = (node *)malloc( 2*NumSamples*sizeof(node) );

	list = (node **)malloc( NumSamples*sizeof(node *) );

	for(in=0; in<2*NumSamples; in++)
	{	
		Ctree[in].time = 0.;
		Ctree[in].desc1 = Ctree[in].desc2 = 0;
		Ctree[in].desc1index = Ctree[in].desc2index = -1;
		Ctree[in].index = in;
	}

	/* set pointers to all leaves */
	for(in=0; in<NumSamples; in++)
		list[in] = &(Ctree[in]);
		/* list[in] = Ctree + in; (HAS) */
		/* however, the latter might be faster (HAS) */

	if(factor == 1.0)
	{
		/* Generate the times of the nodes */
		/*fprintf(stderr,"WF ");*/

		t =0.0;

		/* set time before present for all nodes */
		for( in=NumSamples ; in>1; in--)
		{	
			t += -2.0 * log(1.0-randdouple(MyThreadNum)) /( ((double)in)*(in-1) );
			Ctree[2*NumSamples - in].time = t;
		}

	} else {
		/* Generate the times of the nodes */
		/*fprintf(stderr,"GR ");*/

		rr = - log(factor)/Tt;

		t = 0.;
		s = 0.;

		/* set time before present for all nodes */
		for( in=NumSamples; in>1; in--)
		{	
			t_i =  (-2.0 * log(1.-randdouple(MyThreadNum)) /( ((double)in)*(in-1) ));

			if( s < Tt )
			{
				L_s = 1./rr * ( exp(rr*s) -1. );
	
				t = 1./rr * log( 1 + rr * (t_i+L_s) );

				if( t > Tt)
					t = Tt + factor * ( t_i + L_s - 1./rr*(1./factor-1.) );
			}
			else
				t = s + factor * t_i;

			s = t;
	
			Ctree[2*NumSamples - in].time = 1./factor * t;
		}
	}

	/* Generate the topology of the tree  */
	for( in=NumSamples ; in>1; in--)
	{	
		pick = in*randdouple(MyThreadNum);
		list[pick]->ancestor = Ctree + 2*NumSamples - in;
		Ctree[2*NumSamples - in].desc1 = list[pick];
		Ctree[2*NumSamples - in].desc1index = list[pick]->index;
		list[pick] = list[in-1];
		pick = (in-1)*randdouple(MyThreadNum);
		list[pick]->ancestor = Ctree + 2*NumSamples - in;
		Ctree[2*NumSamples - in].desc2 = list[pick];
		Ctree[2*NumSamples - in].desc2index = list[pick]->index;
		list[pick] = Ctree + 2*NumSamples - in;
		Ctree[2*NumSamples - in].index = 2*NumSamples - in;
	}

	Ctree[2*NumSamples - 2].ancestor = NULL;

	free(list);

#if 0
	C_2_Ttree(Ctree, tree); /* TO BE MADE THREAD-SAFE (HAS) */
#endif
	C_2_Ttree(Ctree, outtreeptr, outTlistptr);

	free(Ctree);

} /* grow_tree */

/*****************************************************************/


/*	These routines are by Ziheng Yang for generating
	gamma distributed random variables. */

double LnGamma (double alpha);
double IncompleteGamma (double x, double alpha, double ln_gamma_alpha);
double PointNormal (double prob);
double PointChi2 (double prob, double v);

	
double rndgamma1 (double s);
double rndgamma2 (double s);


double rndgamma (double s)
{
	double	r=0.0;
	
	if (s <= 0.0)      
		puts ("Error gamma..");
	else if (s < 1.0)  
		r = rndgamma1 (s);
	else if (s > 1.0)  
		r = rndgamma2 (s);
	else           
		r =- log(randdouple(MyThreadNum));
	return (r);
}


double rndgamma1 (double s)
{

	double			r, x=0.0, small=1e-37, w;
	static double	a, p, uf, ss=10.0, d;
	
	if (s!=ss) 
		{
		a  = 1.0-s;
		p  = a/(a+s*exp(-a));
		uf = p*pow(small/a,s);
		d  = a*log(a);
		ss = s;
		}
	for (;;) 
		{
		r = randdouple(MyThreadNum);
		if (r > p)        
			x = a-log((1.0-r)/(1.0-p)), w=a*log(x)-d;
		else if (r>uf)  
			x = a*pow(r/p,1/s), w=x;
		else            
			return (0.0);
		r = randdouple(MyThreadNum);
		if (1.0-r <= w && r > 0.0)
			if (r*(w+1.0) >= 1.0 || -log(r) <= w)  
				continue;
		break;
		}
	return (x);
}


double rndgamma2 (double s)
{

	double			r ,d, f, g, x;
	static double	b, h, ss=0;
	
	if (s!=ss) 
		{
		b  = s-1.0;
		h  = sqrt(3.0*s-0.75);
		ss = s;
		}
	for (;;) 
		{
		r = randdouple(MyThreadNum);
		g = r-r*r;
		f = (r-0.5)*h/sqrt(g);
		x = b+f;
		if (x <= 0.0) 
			continue;
		r = randdouple(MyThreadNum);
		d = 64*r*r*g*g*g;
		if (d*x < x-2.0*f*f || log(d) < 2*(b*log(x/b)-f))  
			break;
		}
	return (x);
}




double LnGamma (double alpha)
{
/* returns ln(gamma(alpha)) for alpha>0, accurate to 10 decimal places.  
   Stirling's formula is used for the central polynomial part of the procedure.
   Pike MC & Hill ID (1966) Algorithm 291: Logarithm of the gamma function.
   Communications of the Association for Computing Machinery, 9:684
*/
   double x=alpha, f=0, z;

   if (x<7) {
      f=1;  z=x-1;
      while (++z<7)  f*=z;
      x=z;   f=-log(f);
   }
   z = 1/(x*x);
   return  f + (x-0.5)*log(x) - x + .918938533204673 
	  + (((-.000595238095238*z+.000793650793651)*z-.002777777777778)*z
	       +.083333333333333)/x;  
}

double IncompleteGamma (double x, double alpha, double ln_gamma_alpha)
{
/* returns the incomplete gamma ratio I(x,alpha) where x is the upper 
	   limit of the integration and alpha is the shape parameter.
   returns (-1) if in error
   ln_gamma_alpha = ln(Gamma(alpha)), is almost redundant.
   (1) series expansion     if (alpha>x || x<=1)
   (2) continued fraction   otherwise
   RATNEST FORTRAN by
   Bhattacharjee GP (1970) The incomplete gamma integral.  Applied Statistics,
   19: 285-287 (AS32)
*/
   int i;
   double p=alpha, g=ln_gamma_alpha;
   double accurate=1e-8, overflow=1e30;
   double factor, gin=0, rn=0, a=0,b=0,an=0,dif=0, term=0, pn[6];

   if (x==0) return (0);
   if (x<0 || p<=0) return (-1);

   factor=exp(p*log(x)-x-g);   
   if (x>1 && x>=p) goto l30;
   /* (1) series expansion */
   gin=1;  term=1;  rn=p;
 l20:
   rn++;
   term*=x/rn;   gin+=term;

   if (term > accurate) goto l20;
   gin*=factor/p;
   goto l50;
 l30:
   /* (2) continued fraction */
   a=1-p;   b=a+x+1;  term=0;
   pn[0]=1;  pn[1]=x;  pn[2]=x+1;  pn[3]=x*b;
   gin=pn[2]/pn[3];
 l32:
   a++;  b+=2;  term++;   an=a*term;
   for (i=0; i<2; i++) pn[i+4]=b*pn[i+2]-an*pn[i];
   if (pn[5] == 0) goto l35;
   rn=pn[4]/pn[5];   dif=fabs(gin-rn);
   if (dif>accurate) goto l34;
   if (dif<=accurate*rn) goto l42;
 l34:
   gin=rn;
 l35:
   for (i=0; i<4; i++) pn[i]=pn[i+2];
   if (fabs(pn[4]) < overflow) goto l32;
   for (i=0; i<4; i++) pn[i]/=overflow;
   goto l32;
 l42:
   gin=1-factor*gin;

 l50:
   return (gin);
}


/* functions concerning the CDF and percentage points of the gamma and
   Chi2 distribution
*/
double PointNormal (double prob)
{
/* returns z so that Prob{x<z}=prob where x ~ N(0,1) and (1e-12)<prob<1-(1e-12)
   returns (-9999) if in error
   Odeh RE & Evans JO (1974) The percentage points of the normal distribution.
   Applied Statistics 22: 96-97 (AS70)

   Newer methods:
     Wichura MJ (1988) Algorithm AS 241: the percentage points of the
       normal distribution.  37: 477-484.
     Beasley JD & Springer SG  (1977).  Algorithm AS 111: the percentage 
       points of the normal distribution.  26: 118-121.

*/
   double a0=-.322232431088, a1=-1, a2=-.342242088547, a3=-.0204231210245;
   double a4=-.453642210148e-4, b0=.0993484626060, b1=.588581570495;
   double b2=.531103462366, b3=.103537752850, b4=.0038560700634;
   double y, z=0, p=prob, p1;

   p1 = (p<0.5 ? p : 1-p);
   if (p1<1e-20) return (-9999);

   y = sqrt (log(1/(p1*p1)));   
   z = y + ((((y*a4+a3)*y+a2)*y+a1)*y+a0) / ((((y*b4+b3)*y+b2)*y+b1)*y+b0);
   return (p<0.5 ? -z : z);
}


double PointChi2 (double prob, double v)
{
/* returns z so that Prob{x<z}=prob where x is Chi2 distributed with df=v
   returns -1 if in error.   0.000002<prob<0.999998
   RATNEST FORTRAN by
       Best DJ & Roberts DE (1975) The percentage points of the 
       Chi2 distribution.  Applied Statistics 24: 385-388.  (AS91)
   Converted into C by Ziheng Yang, Oct. 1993.
*/
   double e=.5e-6, aa=.6931471805, p=prob, g;
   double xx, c, ch, a=0,q=0,p1=0,p2=0,t=0,x=0,b=0,s1,s2,s3,s4,s5,s6;

   if (p<.000002 || p>.999998 || v<=0) return (-1);

   g = LnGamma (v/2);
   xx=v/2;   c=xx-1;
   if (v >= -1.24*log(p)) goto l1;

   ch=pow((p*xx*exp(g+xx*aa)), 1/xx);
   if (ch-e<0) return (ch);
   goto l4;
l1:
   if (v>.32) goto l3;
   ch=0.4;   a=log(1-p);
l2:
   q=ch;  p1=1+ch*(4.67+ch);  p2=ch*(6.73+ch*(6.66+ch));
   t=-0.5+(4.67+2*ch)/p1 - (6.73+ch*(13.32+3*ch))/p2;
   ch-=(1-exp(a+g+.5*ch+c*aa)*p2/p1)/t;
   if (fabs(q/ch-1)-.01 <= 0) goto l4;
   else                       goto l2;
  
l3: 
   x=PointNormal (p);
   p1=0.222222/v;   ch=v*pow((x*sqrt(p1)+1-p1), 3.0);
   if (ch>2.2*v+6)  ch=-2*(log(1-p)-c*log(.5*ch)+g);
l4:
   q=ch;   p1=.5*ch;
   if ((t=IncompleteGamma (p1, xx, g))<0) {
      return (-1);
   }
   p2=p-t;
   t=p2*exp(xx*aa+g+p1-c*log(ch));   
   b=t/ch;  a=0.5*t-b*c;

   s1=(210+a*(140+a*(105+a*(84+a*(70+60*a))))) / 420;
   s2=(420+a*(735+a*(966+a*(1141+1278*a))))/2520;
   s3=(210+a*(462+a*(707+932*a)))/2520;
   s4=(252+a*(672+1182*a)+c*(294+a*(889+1740*a)))/5040;
   s5=(84+264*a+c*(175+606*a))/2520;
   s6=(120+c*(346+127*c))/5040;
   ch+=t*(1+0.5*t*s1-b*c*(s1-b*(s2-b*(s3-b*(s4-b*(s5-b*s6))))));
   if (fabs(q/ch-1) > e) goto l4;

   return (ch);
}


#define PointGamma(prob,alpha,beta) PointChi2(prob,2.0*(alpha))/(2.0*(beta))

int DiscreteGamma (double freqK[], 
		double rK[], 
		double alfa, 
		double beta, 
		int K, 
		int median)
{
/* discretization of gamma distribution with equal proportions in each 
   category
*/
   int i;
   double gap05=1.0/(2.0*K), t, factor, lnga1;
   
   
   if( alfa > 0 )
   {
   factor =alfa/beta*K;

   if (median) {
      for (i=0; i<K; i++) rK[i]=PointGamma((i*2.0+1)*gap05, alfa, beta);
      for (i=0,t=0; i<K; i++) t+=rK[i];
      for (i=0; i<K; i++)     rK[i]*=factor/t;
   }
   else {
      lnga1=LnGamma(alfa+1);
      for (i=0; i<K-1; i++)
	 freqK[i]=PointGamma((i+1.0)/K, alfa, beta);
      for (i=0; i<K-1; i++)
	 freqK[i]=IncompleteGamma(freqK[i]*beta, alfa+1, lnga1);
      rK[0] = freqK[0]*factor;
      rK[K-1] = (1-freqK[K-2])*factor;
      for (i=1; i<K-1; i++)  rK[i] = (freqK[i]-freqK[i-1])*factor;
   }
   for (i=0; i<K; i++) freqK[i]=1.0/K;

/*for (i=0; i<K; i++)
fprintf(stderr,"%f ",rK[i]);
fprintf(stderr,"\n");
*/

   return (0);
   
   }
   
   else
   {
   for (i=0; i<K; i++)     rK[i] = 1;
   for (i=0; i<K; i++) freqK[i]=1.0/K;
   
   return (0); /* ??? (HAS) */
   }
}

/************************************************************/

void nrerror(char error_text[])
/* Numerical Recipes standard error handler */
{
	fprintf(stderr,"Numerical Recipes run-time error...\n");
	fprintf(stderr,"%s\n",error_text);
	fprintf(stderr,"...now exiting to system...\n");
	exit(1);
}


double *dvector(long nl, long nh)
/* allocate a double vector with subscript range v[nl..nh] */
{
	double *v;

	v=(double *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(double)));
	if (!v) nrerror("allocation failure in dvector()");
	return v-nl+NR_END;
}

double ***f3tensor(long nrl, long nrh, long ncl, long nch, long ndl, long ndh)
/* allocate a double 3tensor with range t[nrl..nrh][ncl..nch][ndl..ndh] */
{
	long i,j,nrow=nrh-nrl+1,ncol=nch-ncl+1,ndep=ndh-ndl+1;
	double ***t;

	/* allocate pointers to pointers to rows */
	t=(double ***) malloc((size_t)((nrow+NR_END)*sizeof(double**)));
	if (!t) nrerror("allocation failure 1 in f3tensor()");
	t += NR_END;
	t -= nrl;

	/* allocate pointers to rows and set pointers to them */
	t[nrl]=(double **) malloc((size_t)((nrow*ncol+NR_END)*sizeof(double*)));
	if (!t[nrl]) nrerror("allocation failure 2 in f3tensor()");
	t[nrl] += NR_END;
	t[nrl] -= ncl;

	/* allocate rows and set pointers to them */
	t[nrl][ncl]=(double *) malloc((size_t)((nrow*ncol*ndep+NR_END)*sizeof(double)));
	if (!t[nrl][ncl]) nrerror("allocation failure 3 in f3tensor()");
	t[nrl][ncl] += NR_END;
	t[nrl][ncl] -= ndl;

	for(j=ncl+1;j<=nch;j++) t[nrl][j]=t[nrl][j-1]+ndep;
	for(i=nrl+1;i<=nrh;i++) {
		t[i]=t[i-1]+ncol;
		t[i][ncl]=t[i-1][ncl]+ncol*ndep;
		for(j=ncl+1;j<=nch;j++) t[i][j]=t[i][j-1]+ndep;
	}

	/* return pointer to array of pointers to rows */
	return t;
}

