/*
 * sub1.h
 *
 *
 * Part of (p)IPHULA 1.16 (May 2007)
 *
 * (c) 2004-2007 by Heiko A. Schmidt, Jutta Buschbom, 
 *                  Gunter Weiss, and Arndt von Haeseler
 * (c) 1998-2004 by Gunter Weiss and Arndt von Haeseler
 *
 * All parts of the source except where indicated are distributed 
 * under the GNU public licence 2. 
 * See http://www.opensource.org/licenses/gpl-license.php details.
 *
 * ($Id: sub1.h,v 1.5 2005/09/29 13:29:57 hschmidt Exp $)
 *
 */

#ifndef SUB1_H
#define SUB1_H

#include "tree.h"


typedef struct Node {	double time;		/* time before present (coalescent) */
		int index;		/* node number */
		struct Node *desc1;	/* pointer  to descendent 1 */
		int desc1index;		/* node no. of descendent 1 */
		struct Node *desc2;	/* pointer  to descendent 2 */
		int desc2index;		/* node no. of descendent 2 */
		struct Node *ancestor;	/* pointer  to parent */
	    } node;

extern int NumThreads;	/* number of threads (1, if no OpenMP) (HAS) */
extern int MyThreadNum;	/* thread number (0, if no OpenMP) (HAS) */
extern int NumSamples;	/* number of taxa ??? (HAS) */
extern long NumRuns;	/* simulation runs to perform */
extern TTree *tree;	/* random tree structure for the simulation */
extern TNode **Tlist;	/* tree nodes of simulation trees */
extern double MeanPWDist;

extern double MinTheta;	/* min theta to test */
extern double MaxTheta;	/* max theta to test */
extern double MinTau;	/* min tau to test */
extern double MaxTau;	/* max tau to test */
/* extern double Rho; */
#define MAX_RHO_STEPS 30
extern double Rho_arr[MAX_RHO_STEPS];
extern int Rhosteps;
extern int Thetasteps;	/* number of thetas to test */
extern int Tausteps;	/* number of taus to test */

/*** input dialog routines ***/
void OpenDialog(void);
void SubsProDialog(FILE *infp, FILE *outfp);
void DataDialog(FILE *infp, FILE *outfp);
void SimParDialog(FILE *infp, FILE *outfp);

/*** genrate random trees ***/
/* void grow_tree(double factor, double Tt); */
/* void grow_tree(double factor, double Tt, TTree *tree, TNode **Tlist); */
void grow_tree(double factor, double Tt, TTree **outtreeptr, TNode ***outTlistptr);

/*** genrate random numbers ***/
long seedran(long randseed);	/* set random seed */
double randdouple(int);		/* draw random number [0..1[ */
double rndgamma (double s);	/* draw random gamma (based on randdouple) */

int DiscreteGamma (double freqK[], double rK[], double alfa, double beta, int K, int median);

/*** error output ***/
void nrerror(char error_text[]);

/*** alloc routines ***/
double *dvector(long nl, long nh);
double ***f3tensor(long nrl, long nrh, long ncl, long nch, long ndl, long ndh);

#endif /* SUB1_H */
