/*
 * sub2.c
 *
 *
 * Part of (p)IPHULA 1.16 (May 2007)
 *
 * (c) 2004-2007 by Heiko A. Schmidt, Jutta Buschbom, 
 *                  Gunter Weiss, and Arndt von Haeseler
 * (c) 1998-2004 by Gunter Weiss and Arndt von Haeseler
 *
 * All parts of the source except where indicated are distributed 
 * under the GNU public licence 2. 
 * See http://www.opensource.org/licenses/gpl-license.php details.
 *
 * ($Id: sub2.c,v 1.6 2005/09/29 13:29:57 hschmidt Exp $)
 *
 */

/* these routines are mainly a shortened copy of: */

/*  Nucleotide Sequence Generator - seq-gen, version 1.02  */
/*  (c) Copyright 1996, Andrew Rambaut & Nick Grassly      */
/*  Department of Zoology, University of Oxford            */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#ifdef _OPENMP
#	define USE_OPENMP 1
#	include "omp.h"
#else
#       ifdef USE_OPENMP
#               undef USE_OPENMP
#       endif
#endif

#ifdef MPI
#       define USE_MPI 1
#       include <mpi.h>
#else
#       ifdef USE_MPI
#               undef USE_MPI
#       endif
#endif


#define EXTERN extern
#include "iphula.h"
#include "sub1.h"
#include "sub2.h"
#include "config.h"

FILE *outseq;

int numTaxa, NumBases;

double GammaShape;		/* shape parameter for Gamma-distr. */
int    NumCats;			/* # rate categories (1=homogeneous) */ 
double CatRate[MAX_RATE_CATS];
double freqRate[MAX_RATE_CATS];

short *Categories;

char *nucleotides="ACGT";

double Freq[4], CumulFreq[4];
double freqR, freqY, freqAG, freqCT;
double freqA, freqC, freqG, freqT;
double ksi, kappa;

static double beta, beta_A_R, beta_A_Y;
static double tab1A, tab2A, tab3A;
static double tab1C, tab2C, tab3C;
static double tab1G, tab2G, tab3G;
static double tab1T, tab2T, tab3T;


static double matrix[MAX_RATE_CATS][16];
/* static double vector[4]; (HAS) */
/* static double *siteRates; (HAS) */


/* prototypes */

char SetBase(double *P);
void RandomSequence(char *seq);
void MutateSequence(char *seq, double len);
void EvolveNode(TNode *anc, TNode *des, double scale);



/*************************************/
/* TO BE MADE THREAD-SAFE (HAS) */
/* external  variables: Freq, CumulFreq, freqR, freqY, 
			freqA, freqC, freqG, freqT, 
			tab1A, tab2A, tab3A, 
			tab1C, tab2C, tab3C, 
			tab1G, tab2G, tab3G, 
			tab1T, tab2T, tab3T, 
			kappa, ksi, beta
			beta_A_R, beta_A_Y
			*/
/* internal  variables: i,j,k, alpha1, alpha2, sumFreq, freqAG, freqCT */
/* parameter variables: */
/* calling:    */
/* called by: main */
void SetModel(void)
{	

	/* int i,j,k; (HAS) */
	double alpha1, alpha2, sumFreq;
	double freqAG, freqCT;


	sumFreq=Freq[0]+Freq[1]+Freq[2]+Freq[3];
	if (sumFreq!=1.0) {
		Freq[0]*=1.0/sumFreq;
		Freq[1]*=1.0/sumFreq;
		Freq[2]*=1.0/sumFreq;
		Freq[3]*=1.0/sumFreq;
	}
	freqA=Freq[0];
	freqC=Freq[1];
	freqG=Freq[2];
	freqT=Freq[3];
	CumulFreq[0]=freqA;
	CumulFreq[1]=CumulFreq[0]+freqC;
	CumulFreq[2]=CumulFreq[1]+freqG;
	CumulFreq[3]=CumulFreq[2]+freqT;

	freqR=freqA+freqG;
	freqY=freqC+freqT;
	freqAG=freqA*freqG;
	freqCT=freqC*freqT;
		
	tab1A=freqA*freqY/freqR;
	tab2A=freqG/freqR;
	tab3A=freqA/freqR;
	tab1C=freqC*freqR/freqY;
	tab2C=freqT/freqY;
	tab3C=freqC/freqY;
	tab1G=freqG*freqY/freqR;
	tab2G=freqA/freqR;
	tab3G=freqG/freqR;
	tab1T=freqT*freqR/freqY;
	tab2T=freqC/freqY;
	tab3T=freqT/freqY;


	/* eigenwerte */	

	alpha1 = 4 * kappa / (ksi+1.);
	alpha2 = ksi * alpha1;			

	beta=-1.0/(2*(freqR*freqY + alpha1*freqAG + alpha2*freqCT)); /* nr 1*/
			
	beta_A_R=beta*(alpha1*freqR + freqY);	/* nr 2*/

	beta_A_Y=beta*(alpha2*freqY + freqR);	/* nr 3*/
		

} /* SetModel */



/* TO BE MADE THREAD-SAFE (HAS) */
/* constant external  variables: (set in SetModel)
			beta, beta_A_R, beta_A_Y,
			freqA, freqC, freqG, freqT, 
			tab1A, tab2A, tab3A, 
			tab1C, tab2C, tab3C, 
			tab1G, tab2G, tab3G, 
			tab1T, tab2T, tab3T */
/* internal  variables: aa, bbR, bbY, i, j, k, expt, P */
/* parameter variables: len, matrix */
/* external  constants: PIJ_TS_A, PIJ_TS_C, PIJ_TS_G, PIJ_TS_T,
			PIJ_TV_A, PIJ_TV_C, PIJ_TV_G, PIJ_TV_T,
			PIJ_SAME_A, PIJ_SAME_C, PIJ_SAME_G, PIJ_SAME_T */
/* calling:   exp */
/* called by: MutateSequence */
void SetMatrix(double *matrix, double len)
{	

	double aa, bbR, bbY;
	/* int i,j,k; (HAS) */
	/* double expt[4]; (HAS) */
	/* double *P; (HAS) */

	aa=exp(beta*len);
	bbR=exp(beta_A_R*len);
	bbY=exp(beta_A_Y*len);
		
	matrix[0] = freqA+(tab1A*aa)+(tab2A*bbR);	/* PIJ_SAME_A (HAS) */
	matrix[1] = freqC*(1-aa);			/* PIJ_TV_C (HAS) */
	matrix[2] = freqG+(tab1G*aa)-(tab3G*bbR);	/* PIJ_TS_G (HAS) */
	matrix[3] = freqT*(1-aa);			/* PIJ_TV_T (HAS) */

	matrix[4] = freqA*(1-aa);			/* PIJ_TV_A (HAS) */
	matrix[5] = freqC+(tab1C*aa)+(tab2C*bbY);	/* PIJ_SAME_C (HAS) */
	matrix[6] = freqG*(1-aa);			/* PIJ_TV_G (HAS) */
	matrix[7] = freqT+(tab1T*aa)-(tab3T*bbY);	/* PIJ_TS_T (HAS) */

	matrix[8] = freqA+(tab1A*aa)-(tab3A*bbR);	/* PIJ_TS_A (HAS) */
	matrix[9] = matrix[1];				/* PIJ_TV_C */
	matrix[10]= freqG+(tab1G*aa)+(tab2G*bbR);	/* PIJ_SAME_G (HAS) */
	matrix[11]= matrix[3];				/* PIJ_TV_T */

	matrix[12]= matrix[4];				/* PIJ_TV_A */
	matrix[13]= freqC+(tab1C*aa)-(tab3C*bbY);	/* PIJ_TS_C (HAS) */
	matrix[14]= matrix[6];				/* PIJ_TV_G */
	matrix[15]= freqT+(tab1T*aa)+(tab2T*bbY);	/* PIJ_SAME_T (HAS) */



/* the rows are cumulative to help with picking one using
   a random number */

	matrix[1] += matrix[0];
	matrix[2] += matrix[1];
	matrix[3] += matrix[2]; /* This should always be 1.0... */

	matrix[5] += matrix[4];
	matrix[6] += matrix[5];
	matrix[7] += matrix[6]; /* ...but it is easier to spot bugs... */
	
	matrix[9] += matrix[8];
	matrix[10]+= matrix[9];
	matrix[11]+= matrix[10]; /* ...though less efficient... */
	
	matrix[13]+= matrix[12];
	matrix[14]+= matrix[13];
	matrix[15]+= matrix[14]; /* ...but probably not much. */
} /* SetMatrix */


void SetVector(double *vector, short base, double len)
{
	double aa, bbR, bbY;
	/* int i,j,k; (HAS) */
	/* double expt[4]; (HAS) */
	/* double *P; (HAS) */


	aa=exp(beta*len);
	bbR=exp(beta_A_R*len);
	bbY=exp(beta_A_Y*len);

		
		switch (base) {
			case 0:
				vector[0]=freqA+(tab1A*aa)+(tab2A*bbR);			/* PIJ_SAME_A (HAS) */
				vector[1]=freqC*(1-aa)                +vector[0];	/* PIJ_TV_C (HAS) */
				vector[2]=freqG+(tab1G*aa)-(tab3G*bbR)+vector[1];	/* PIJ_TS_G (HAS) */
				vector[3]=freqT*(1-aa)                +vector[2];	/* PIJ_TV_T (HAS) */
			break;
			case 1:
				vector[0]=freqA*(1-aa);					/* PIJ_TV_A (HAS) */
				vector[1]=freqC+(tab1C*aa)+(tab2C*bbY)+vector[0];	/* PIJ_SAME_C (HAS) */
				vector[2]=freqG*(1-aa)                +vector[1];	/* PIJ_TV_G (HAS) */
				vector[3]=freqT+(tab1T*aa)-(tab3T*bbY)+vector[2];	/* PIJ_TS_T (HAS) */
			break;
			case 2:
				vector[0]=freqA+(tab1A*aa)-(tab3A*bbR);			/* PIJ_TS_A (HAS) */
				vector[1]=freqC*(1-aa)                +vector[0];	/* PIJ_TV_C (HAS) */
				vector[2]=freqG+(tab1G*aa)+(tab2G*bbR)+vector[1];	/* PIJ_SAME_G (HAS) */
				vector[3]=freqT*(1-aa)                +vector[2];	/* PIJ_TV_T (HAS) */
			break;
			case 3:
				vector[0]=freqA*(1-aa);					/* PIJ_TV_A (HAS) */
				vector[1]=freqC+(tab1C*aa)-(tab3C*bbY)+vector[0];	/* PIJ_TS_C (HAS) */
				vector[2]=freqG*(1-aa)                +vector[1];	/* PIJ_TV_G (HAS) */
				vector[3]=freqT+(tab1T*aa)+(tab2T*bbY)+vector[2];	/* PIJ_SAME_T (HAS) */
			break;
		}
} /* SetVector */



/*************************************/
/* TO BE MADE THREAD-SAFE (HAS) */
/* external  variables: */
/* internal  variables: P */
/* parameter variables: n */
/* calling:   fprintf, exit, calloc */
/* called by:  */
void *CAllocMem(long n)
{
	void *P;
	
	if ( (P=calloc(n, 1))==NULL ) {
		fprintf(stderr, "Out of memory allocating\n");
		exit(0);
	}
	
	return P;
} /* CAllocMem */


/* functions */

/* TO BE MADE THREAD-SAFE (HAS) */
/* external  variables: NumBases */
/* internal  variables: P */
/* parameter variables: tree */
/* calling:   CAllocMem */
/* called by: seq_gen */
void CreateSequences(TTree *tree)
{
	TNode *P;
	
	P=tree->nodeList;
	while (P!=NULL) {
		P->sequence=CAllocMem(NumBases+1);	/* TO BE MADE THREAD-SAFE (HAS) */
		P=P->next;
	}
} /* CreateSequences */


/* TO BE MADE THREAD-SAFE (HAS) */
/* new external  variables: NumBases */
/* new internal  variables: i */
/* new parameter variables: NumCats, Categories */
/* new calling:    */
/* new called by: seq_gen */
/* old external  variables: NumCats, NumBases, Categories */
/* old internal  variables: i */
/* old parameter variables: */
/* old calling:    */
/* old called by: seq_gen */
void SetCategories(int NumCats, short *Categories)
{
	int i;

	if(NumCats>1) { /* rate heterogeneity */
		for (i=0; i<NumBases; i++)
			Categories[i]=(int)(randdouple(MyThreadNum)*NumCats);
	} else { /* rate homogeneity */
		for (i=0; i<NumBases; i++)
			Categories[i]= 0;
	}
		
} /* SetCategories */


/* TO BE MADE THREAD-SAFE (HAS) */
/* external  variables: */
/* internal  variables: j, r */
/* parameter variables: cumulfreq */
/* calling:   rand2 */
/* called by: RandomSequence, MutateSequence */
char SetBase(double *cumulfreq)
{
	char j;
	double r;
	
	r=randdouple(MyThreadNum);
	for (j='\0'; r>(*cumulfreq) && j<'\3'; j++) 
		cumulfreq++; /* next nucl. freq. */
	return j;
} /* SetBase */


/* TO BE MADE THREAD-SAFE (HAS) */
/* external  variables: NumBases, CumulFreq */
/* internal  variables: i, P */
/* parameter variables: seq */
/* calling:   SetBase */
/* called by: EvolveSequences */
void RandomSequence(char *seq)
{
	int i;
	char *P;
	
	P=seq;
	for (i=0; i<NumBases; i++) {
		*P=SetBase(CumulFreq);	/* TO BE MADE THREAD-SAFE (HAS) */
		P++;
	}
	*P='\0';
} /* RandomSequence */

/* TO BE MADE THREAD-SAFE (HAS) */
/* external  variables: NumCats, matrix, CatRate, NumBases, Categories */
/* internal  variables: i, cat, Q, R, P */
/* parameter variables: seq, len */
/* calling:   SetMatrix, SetBase */
/* called by: EvolveNode */
void MutateSequence(char *seq, double len)
{
	int i;		/* nucleotide loop counter */
	/* int cat; (HAS) */
	/* double *Q; (HAS) */
	short *R;
	char *P;
	
	P=seq;
	

	for (i=0; i<NumCats; i++)
		SetMatrix(matrix[i], CatRate[i]*len);	/* TO BE MADE THREAD-SAFE (HAS) */
			
	R=Categories;
	for (i=0; i<NumBases; i++) {
		*P=SetBase(matrix[*R]+(*P<<2));	/* TO BE MADE THREAD-SAFE (HAS) */
		P++;
		R++;
	}
	
} /* MutateSequence */


/* TO BE MADE THREAD-SAFE (HAS) */
/* external  variables: NumBases */
/* internal  variables: len */
/* parameter variables: anc, des, scale */
/* calling:   SetBase, MutateSequence, memcpy, EvolveNode */
/* called by: EvolveNode, EvolveSequences */
void EvolveNode(TNode *anc, TNode *des, double scale)
{
	double len;
	
	if (scale<0.0)
		len=des->length0;
	else
		len=des->length0*scale;
		
	memcpy(des->sequence, anc->sequence, NumBases);

	MutateSequence(des->sequence, len);	/* TO BE MADE THREAD-SAFE (HAS) */
	
	if (des->tipNo==-1) {
		EvolveNode(des, des->branch1, scale);
		EvolveNode(des, des->branch2, scale);
	}
} /* EvolveNode */


/* TO BE MADE THREAD-SAFE (HAS) */
/* external  variables: */
/* internal  variables: */
/* parameter variables: tree, scale */
/* calling:   EvolveNode, RandomSequence */
/* called by: seq_gen */
void EvolveSequences(TTree *tree, double scale)
{
	RandomSequence(tree->root->sequence);			/* TO BE MADE THREAD-SAFE (HAS) */
	
	EvolveNode(tree->root, tree->root->branch1, scale);	/* TO BE MADE THREAD-SAFE (HAS) */
	EvolveNode(tree->root, tree->root->branch2, scale);
	if (!tree->rooted)
		EvolveNode(tree->root, tree->root->branch0, scale);

} /* EvolveSequences */


void OutSeq(TTree *tree)
{
	int i,j;
	char *P1;

	if ( (outseq = fopen("sequences","a")) == NULL)
		fprintf(stderr,"seqfile-problems");


	for(i=0; i < numTaxa; i++)
	{
		P1=tree->tips[i]->sequence;
		fprintf(outseq,"seq%-7d",i);

		for (j=0; j<NumBases; j++) 
		{
			fprintf(outseq,"%c",*P1);
			P1++;
		}

		fprintf(outseq,"\n");
	}

	fclose(outseq);

}




/* TO BE MADE THREAD-SAFE (HAS) */
/* new external  variables: numTaxa, NumBases */
/* new internal  variables: s1, s2, j, var, obs_pwd, diff, P1, P2, P */
/* new parameter variables: tree, obs_numvarpos, obs_meanpwdist*/
/* new calling:    */
/* new called by: seq_gen */
/* old external  variables: numTaxa, NumBases, observed */
/* old internal  variables: s1, s2, j, var, obs_pwd, diff, P1, P2, P */
/* old parameter variables: tree */
/* old calling:    */
/* old called by: seq_gen */
void Stats(TTree *tree, int *obs_numvarpos, double *obs_meanpwdist)
{
	int s1, s2, j, var;
	double obs_pwd;
	/* double diff; (HAS) */
	char *P1, *P2;

	/* TNode *P; (HAS) */


	obs_pwd = 0.0;

	for (s1=0; s1<numTaxa-1; s1++) 
	{
		for (s2=s1+1; s2<numTaxa; s2++) 
		{
			P1=tree->tips[s1]->sequence;
			P2=tree->tips[s2]->sequence;
				
			for (j=0; j<NumBases; j++) 
			{
				if( (int)(*P1) != (int)(*P2) )
					obs_pwd += 1.0;
		
				P1++;
				P2++;	

				/* TODO: change to: (HAS) */
#if 0
				if( P1[j] != P2[j] ) obs_pwd += 1.0;
#endif
			}
		
		}
	}


	var = 0;

	for (j=0; j<NumBases; j++) 
	{
		P1=tree->tips[0]->sequence+j;

		for (s2=1; s2<numTaxa; s2++) 
		{
			P2=tree->tips[s2]->sequence+j;
	
			if( (int)(*P1) != (int)(*P2) )
			{
				var += 1;
				s2 = tree->numTips;
			}
		}
	}
			/* TODO: change to: (HAS) */
#if 0
	for (j=0; j<NumBases; j++) {
		char c;
		c=tree->tips[0]->sequence[j];
		for (s2=1; s2<numTaxa; s2++) {
			if( c != tree->tips[s2]->sequence[j] ) {
				var ++;
				s2 = numTaxa /* tree->numTips */;
			}
		}
	}
#endif

	*obs_numvarpos  = var;
	*obs_meanpwdist = obs_pwd*2.0/(numTaxa-1.0)/numTaxa;
	/* observed[OBS_NUMVARPOS] = (double)var; */
	/* observed[OBS_MEANPWDIST] = obs_pwd*2.0/(numTaxa-1.0)/numTaxa; */

} /* Stats */


/* TO BE MADE THREAD-SAFE (HAS) - done */
/* new external  variables: numTaxa, NumBases, observed, NumVarPos, MeanPWDist, tolerance */
/* new internal  variables: P, scale */
/* new parameter variables: TTree *tree, double theta0, uli count_positive */
/* new calling: CreateSequences, SetCategories, EvolveSequences */
/* new called by: main */
/* old external  variables: numTaxa, NumBases, observed, NumVarPos, MeanPWDist, tolerance, prob_com */
/* old internal  variables: P, scale */
/* old parameter variables: TTree *tree, int d, int f, int tt, double theta0 */
/* old calling: CreateSequences, SetCategories, EvolveSequences */
/* old called by: main */
/* void seq_gen(TTree *tree, int d, int f, int tt, double theta0, unsigned long int *count_positive, FILE *raw_fp) */
void seq_gen(TTree *tree, double theta0, unsigned long int *count_positive, FILE *raw_fp)
{
	double scale;
	TNode *P;
	int obs_numvarpos;
	double obs_meanpwdist;
 	
	numTaxa=tree->numTips;		/* why not NumSamples ??? (HAS) */

	scale = theta0/2.0/(double)NumBases;

	CreateSequences(tree);			/* TO BE MADE THREAD-SAFE (HAS) */

	SetCategories(NumCats, Categories);	/* TO BE MADE THREAD-SAFE (HAS) */
						
	EvolveSequences(tree, scale);		/* TO BE MADE THREAD-SAFE (HAS) */

/*
OutSeq(tree);
*/
	/* Stats(tree);	*/			/* TO BE MADE THREAD-SAFE (HAS) */
	Stats(tree, &obs_numvarpos, &obs_meanpwdist);

	/* it the sample 'equal' to 'reality' -> counter++ */
	/* if( ((int)observed[OBS_NUMVARPOS]==NumVarPos) && (fabs(observed[OBS_MEANPWDIST]-MeanPWDist)<tolerance) ) */
/* change ;-) */
	/*	fprintf(bla_fp, "\t%d\t%.4f", Rho, Theta, Tau); */
	/*	fprintf(bla_fp, "\t%d,%.4f", obs_numvarpos, obs_meanpwdist); */

	if (writeraw_optn) {
		fprintf(raw_fp, "\t%d,%.4f", obs_numvarpos, obs_meanpwdist);
		fflush(raw_fp);
	}

	if( (obs_numvarpos == NumVarPos) && (fabs(obs_meanpwdist - MeanPWDist) < tolerance) )
		(*count_positive)++;
		/* prob_com[d][f][tt] += 1;	*/

	P=tree->nodeList;
	while (P!=NULL) {
		free(P->sequence);
		P=P->next;
	}
} /* seq_gen */
